# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from __future__ import print_function
from builtins import object
from twisted.internet.defer import Deferred
from twisted.internet.endpoints import TCP4ClientEndpoint, TCP4ServerEndpoint
from twisted.internet.protocol import ClientFactory, Factory

from txi2p.address import I2PAddress
from txi2p.bob.protocol import (I2PClientTunnelCreatorBOBClient,
                                I2PServerTunnelCreatorBOBClient,
                                I2PTunnelRemoverBOBClient,
                                I2PClientTunnelProtocol,
                                I2PServerTunnelProtocol,
                                I2PListeningPort)


class BOBI2PClientFactory(ClientFactory):
    protocol = I2PClientTunnelCreatorBOBClient
    bobProto = None
    canceled = False
    removeTunnelWhenFinished = True

    def _cancel(self, d):
        self.bobProto.sender.transport.abortConnection()
        self.canceled = True

    def __init__(self, reactor, clientFactory, bobEndpoint, dest,
                 tunnelNick=None,
                 inhost='localhost',
                 inport=None,
                 options={}):
        self._reactor = reactor
        self._clientFactory = clientFactory
        self._bobEndpoint = bobEndpoint
        self.dest = dest
        self.tunnelNick = tunnelNick
        self.inhost = inhost
        self.inport = inport
        self.options = options
        self.deferred = Deferred(self._cancel);

    def buildProtocol(self, addr):
        proto = self.protocol()
        proto.factory = self
        self.bobProto = proto
        return proto

    def bobConnectionFailed(self, reason):
        if not self.canceled:
            self.deferred.errback(reason)

    # This method is not called if an endpoint deferred errbacks
    def clientConnectionFailed(self, connector, reason):
        self.bobConnectionFailed(reason)

    def i2pTunnelCreated(self):
        # BOB is now listening for a tunnel.
        # BOB only listens on TCP4 (for now).
        clientEndpoint = TCP4ClientEndpoint(self._reactor, self.inhost, self.inport)
        # Wrap the client Factory.
        wrappedFactory = BOBClientFactoryWrapper(self._clientFactory,
                                                 self._bobEndpoint,
                                                 I2PAddress(self.localDest),
                                                 self.tunnelNick,
                                                 self.removeTunnelWhenFinished)
        wrappedFactory.setDest(self.dest)
        d = clientEndpoint.connect(wrappedFactory)
        def checkProto(proto):
            if proto is None:
                self.deferred.cancel()
            return proto
        d.addCallback(checkProto)
        # When the Deferred returns an IProtocol, pass it on.
        d.chainDeferred(self.deferred)


class BOBI2PServerFactory(Factory):
    protocol = I2PServerTunnelCreatorBOBClient
    bobProto = None
    canceled = False
    removeTunnelWhenFinished = True

    def _cancel(self, d):
        self.bobProto.sender.transport.abortConnection()
        self.canceled = True

    def __init__(self, reactor, serverFactory, bobEndpoint, keyfile,
                 tunnelNick=None,
                 outhost='localhost',
                 outport=None,
                 options={}):
        self._reactor = reactor
        self._serverFactory = serverFactory
        self._bobEndpoint = bobEndpoint
        self._keyfile = keyfile
        self._writeKeypair = False
        self.tunnelNick = tunnelNick
        self.outhost = outhost
        self.outport = outport
        self.options = options
        self.deferred = Deferred(self._cancel)

    def startFactory(self):
        try:
            f = open(self._keyfile, 'r')
            self.keypair = f.read()
            f.close()
        except IOError:
            self.keypair = None
            self._writeKeypair = True

    def buildProtocol(self, addr):
        proto = self.protocol()
        proto.factory = self
        self.bobProto = proto
        return proto

    def bobConnectionFailed(self, reason):
        if not self.canceled:
            self.deferred.errback(reason)

    # This method is not called if an endpoint deferred errbacks
    def clientConnectionFailed(self, connector, reason):
        self.bobConnectionFailed(reason)

    def i2pTunnelCreated(self):
        if self._writeKeypair:
            try:
                f = open(self._keyfile, 'w')
                f.write(self.keypair)
                f.close()
            except IOError:
                print('Could not save keypair')
        # BOB will now forward data to a listener.
        # BOB only forwards to TCP4 (for now).
        serverEndpoint = TCP4ServerEndpoint(self._reactor, self.outport)
        # Wrap the server Factory.
        wrappedFactory = BOBServerFactoryWrapper(self._serverFactory,
                                                 self._bobEndpoint,
                                                 I2PAddress(self.localDest),
                                                 self.tunnelNick,
                                                 self.removeTunnelWhenFinished)
        d = serverEndpoint.listen(wrappedFactory)
        def handlePort(port):
            if port is None:
                self.deferred.cancel()
            serverAddr = I2PAddress(self.localDest)
            p = I2PListeningPort(port, wrappedFactory, serverAddr)
            return p
        d.addCallback(handlePort)
        # When the Deferred returns an IListeningPort, pass it on.
        d.chainDeferred(self.deferred)


class BOBFactoryWrapperCommon(object):
    def __init__(self, wrappedFactory,
                       bobEndpoint,
                       localAddr,
                       tunnelNick,
                       removeTunnelWhenFinished):
        self.w = wrappedFactory
        self.bobEndpoint = bobEndpoint
        self.localAddr = localAddr
        self.tunnelNick = tunnelNick
        self.removeTunnelWhenFinished = removeTunnelWhenFinished

    def __getattr__(self, attr):
        return getattr(self.w, attr)


class BOBClientFactoryWrapper(BOBFactoryWrapperCommon):
    protocol = I2PClientTunnelProtocol

    def setDest(self, dest):
        self.dest = dest

    def buildProtocol(self, addr):
        wrappedProto = self.w.buildProtocol(addr)
        proto = self.protocol(wrappedProto, self.localAddr, self.dest)
        proto.factory = self
        return proto

    def i2pConnectionLost(self, wrappedProto, reason):
        if self.removeTunnelWhenFinished:
            # Notify the underlying Protocol once the tunnel has
            # been removed, in case they stop the reactor.
            rmTunnelFac = BOBI2PClientTunnelRemoverFactory(self.tunnelNick,
                                                           wrappedProto,
                                                           reason)
            self.bobEndpoint.connect(rmTunnelFac)
        else:
            # Notify the underlying Protocol now.
            wrappedProto.connectionLost(reason)


class BOBServerFactoryWrapper(BOBFactoryWrapperCommon):
    protocol = I2PServerTunnelProtocol

    def buildProtocol(self, addr):
        wrappedProto = self.w.buildProtocol(addr)
        proto = self.protocol(wrappedProto, self.localAddr)
        proto.factory = self
        return proto

    def stopListening(self, wrappedPort):
        if self.removeTunnelWhenFinished:
            # Notify the underlying ListeningPort once the tunnel
            # has been removed, in case they stop the reactor.
            rmTunnelFac = BOBI2PServerTunnelRemoverFactory(self.tunnelNick,
                                                           wrappedPort)
            self.bobEndpoint.connect(rmTunnelFac)
        else:
            # Notify the underlying ListeningPort now.
            wrappedPort.stopListening()


class BOBI2PClientTunnelRemoverFactory(ClientFactory):
    protocol = I2PTunnelRemoverBOBClient

    def __init__(self, tunnelNick, protoToNotify, reason):
        self.tunnelNick = tunnelNick
        self._protoToNotify = protoToNotify
        self._reason = reason

    def i2pTunnelRemoved(self):
        # Now that the I2P tunnel has been removed,
        # notify the underlying Protocol.
        self._protoToNotify.connectionLost(self._reason)


class BOBI2PServerTunnelRemoverFactory(ClientFactory):
    protocol = I2PTunnelRemoverBOBClient

    def __init__(self, tunnelNick, portToNotify):
        self.tunnelNick = tunnelNick
        self._portToNotify = portToNotify

    def i2pTunnelRemoved(self):
        # Now that the I2P tunnel has been removed,
        # notify the underlying ListeningPort.
        self._portToNotify.stopListening()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
try:
    # Python 3
    from unittest.mock import Mock
except:
    # Python 2 (library)
    from mock import Mock
from twisted.internet import defer
from twisted.internet.error import ConnectionLost, ConnectionRefusedError
from twisted.internet.protocol import ClientFactory
from twisted.python import failure
from twisted.test import proto_helpers

from txi2p.sam import constants as c

connectionLostFailure = failure.Failure(ConnectionLost())
connectionRefusedFailure = failure.Failure(ConnectionRefusedError())


class SAMProtocolTestMixin(object):
    def makeProto(self, *a, **kw):
        protoClass = kw.pop('_protoClass', self.protocol)
        fac = ClientFactory(*a, **kw)
        fac.nickname = 'foo'
        fac.privKey = None
        fac.port = None
        fac.localPort = None
        fac.options = {}
        fac.sigType = None
        fac.protocol = protoClass
        fac.resultNotOK = Mock()
        def raise_(reason):
            raise reason.value
        fac.connectionFailed = lambda reason: raise_(reason)
        proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return fac, proto

    def test_initSendsHello(self):
        fac, proto = self.makeProto()
        self.assertSubstring('HELLO VERSION', str(proto.transport.value()))

    def test_helloReturnsError(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_pingReceived(self):
        fac, proto = self.makeProto()
        self.addCleanup(proto.receiver.stopPinging)
        proto.transport.clear()
        # Enable keepalive
        proto.receiver.currentRule = 'State_keepalive'
        proto._parser._setupInterp()
        proto.dataReceived(b'PING\n')
        self.assertEquals(
            b'PONG\n',
            proto.transport.value())

    def test_pingReceivedWithData(self):
        fac, proto = self.makeProto()
        self.addCleanup(proto.receiver.stopPinging)
        proto.transport.clear()
        # Enable keepalive
        proto.receiver.currentRule = 'State_keepalive'
        proto._parser._setupInterp()
        proto.dataReceived(b'PING some random data\n')
        self.assertEquals(
            b'PONG some random data\n',
            proto.transport.value())

    def test_pingReceivedResetsTimeout(self):
        fac, proto = self.makeProto()
        self.addCleanup(proto.receiver.stopPinging)
        proto.transport.clear()
        # Enable keepalive
        proto.receiver.currentRule = 'State_keepalive'
        proto._parser._setupInterp()
        proto.receiver._sendPing()
        self.assertEquals(
            'PING %s\n' % proto.receiver.lastPing,
            proto.transport.value().decode('utf-8'))
        self.assertTrue(proto.receiver.pingTimeout.active())
        proto.transport.clear()
        proto.dataReceived(b'PING\n')
        self.assertEquals(
            b'PONG\n',
            proto.transport.value())
        self.assertFalse(proto.receiver.pingTimeout.active())

    def test_validPongResponseResetsTimeout(self):
        fac, proto = self.makeProto()
        self.addCleanup(proto.receiver.stopPinging)
        proto.transport.clear()
        # Enable keepalive
        proto.receiver.currentRule = 'State_keepalive'
        proto._parser._setupInterp()
        proto.receiver._sendPing()
        self.assertEquals(
            'PING %s\n' % proto.receiver.lastPing,
            proto.transport.value().decode('utf-8'))
        self.assertTrue(proto.receiver.pingTimeout.active())
        proto.transport.clear()
        proto.dataReceived(('PONG %s\n' % proto.receiver.lastPing).encode('utf-8'))
        self.assertFalse(proto.receiver.pingTimeout.active())

    def test_invalidPongResponseDoesNotResetTimeout(self):
        fac, proto = self.makeProto()
        self.addCleanup(proto.receiver.stopPinging)
        proto.transport.clear()
        # Enable keepalive
        proto.receiver.currentRule = 'State_keepalive'
        proto._parser._setupInterp()
        proto.receiver._sendPing()
        self.assertEquals(
            'PING %s\n' % proto.receiver.lastPing,
            proto.transport.value().decode('utf-8'))
        self.assertTrue(proto.receiver.pingTimeout.active())
        proto.transport.clear()
        proto.dataReceived(b'PONG not what was expected\n')
        self.assertTrue(proto.receiver.pingTimeout.active())


class SAMFactoryTestMixin(object):
    def setUp(self):
        self.aborted = []

    def makeProto(self, *a, **kw):
        fac = self.factory(*a, **kw)
        proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: self.aborted.append(True)
        proto.makeConnection(transport)
        return fac, proto

    def test_cancellation(self):
        fac, proto = self.makeProto(*self.blankFactoryArgs)
        fac.deferred.cancel()
        self.assert_(self.aborted)
        return self.assertFailure(fac.deferred, defer.CancelledError)

    def test_cancellationBeforeFailure(self):
        fac, proto = self.makeProto(*self.blankFactoryArgs)
        fac.deferred.cancel()
        proto.connectionLost(connectionLostFailure)
        self.assert_(self.aborted)
        return self.assertFailure(fac.deferred, defer.CancelledError)

    def test_cancellationAfterFailure(self):
        fac, proto = self.makeProto(*self.blankFactoryArgs)
        proto.connectionLost(connectionLostFailure)
        fac.deferred.cancel()
        self.assertFalse(self.aborted)
        return self.assertFailure(fac.deferred, ConnectionLost)

    def test_clientConnectionFailed(self):
        fac, proto = self.makeProto(*self.blankFactoryArgs)
        fac.clientConnectionFailed(None, connectionRefusedFailure)
        return self.assertFailure(fac.deferred, ConnectionRefusedError)

    def test_resultNotOK(self):
        fac, proto = self.makeProto(*self.blankFactoryArgs)
        for result, error in list(c.samErrorMap.items()):
            self.assertRaises(error, fac.resultNotOK, result, '')



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

# General I2P grammar
i2pGrammarSource = r"""
digit = anything:x ?(x in '0123456789')
number = <digit+>:ds -> int(ds)
b64char = :x ?(x in 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-~') -> x
b64 = <b64char+>
"""

# BOB grammar
bobGrammarSource = i2pGrammarSource + r"""
KEYS    = b64:keys
KEY     = b64:pubkey
ERROR   = 'ERROR ' <(~'\n' anything)*>:desc '\n' -> (False, desc)
OK      = 'OK ' <(~'\n' anything)*>:info '\n' -> (True, info)
OK_KEY  = 'OK ' KEY:pubkey '\n' -> (True, pubkey)
OK_KEYS = 'OK ' KEYS:keys '\n' -> (True, keys)
DATA    = 'DATA ' <(~'\n' anything)*>:data '\n' -> data

TUNNEL_STATUS = 'NICKNAME: ' <(~' ' anything)*>:nickname ' STARTING: ' <'true'|'false'>:starting ' RUNNING: ' <'true'|'false'>:running ' STOPPING: ' <'true'|'false'>:stopping ' KEYS: ' <'true'|'false'>:keys  ' QUIET: ' <'true'|'false'>:quiet ' INPORT: ' <'not_set'|number>:inport ' INHOST: ' <(~' ' anything)*>:inhost ' OUTPORT: ' <'not_set'|number>:outport ' OUTHOST: ' <(~'\n' anything)*>:outhost '\n' -> {
    'nickname': nickname,
    'starting': starting=='true',
    'running': running=='true',
    'stopping': stopping=='true',
    'keys': keys=='true',
    'quiet': quiet=='true',
    'inport': None if inport=='not_set' else int(inport),
    'inhost': inhost,
    'outport': None if outport=='not_set' else int(outport),
    'outhost': outhost
    }

DATA_TUNNEL_STATUS = 'DATA ' TUNNEL_STATUS:status -> status
OK_TUNNEL_STATUS = 'OK ' TUNNEL_STATUS:status -> (True, status)
OK_DATA_TUNNEL_STATUS = 'OK ' DATA_TUNNEL_STATUS:status -> (True, status)

versionString = <digit+ '.' digit+ '.' digit+>
BOB_init      = 'BOB ' versionString:version '\nOK\n' -> version

BOB_clear     = (ERROR | OK)
BOB_getdest   = (ERROR | OK_KEY)
BOB_getkeys   = (ERROR | OK_KEYS)
BOB_getnick   = (ERROR | OK)
BOB_inhost    = (ERROR | OK)
BOB_inport    = (ERROR | OK)
BOB_list      = ((ERROR:(result, info)                         -> (result, info, []))
                |((DATA_TUNNEL_STATUS)*:data OK:(result, info) -> (result, info, data)))
BOB_newkeys   = (ERROR | OK_KEY)
BOB_option    = (ERROR | OK)
BOB_outhost   = (ERROR | OK)
BOB_outport   = (ERROR | OK)
BOB_quiet     = (ERROR | OK)
BOB_quit      = (OK)
BOB_setkeys   = (ERROR | OK)
BOB_setnick   = (ERROR | OK)
BOB_show      = ((ERROR:(result, info)              -> (result, info, {}))
                |(OK_TUNNEL_STATUS:(result, status) -> (result, '', status)))
BOB_showprops = (ERROR | OK)
BOB_start     = (ERROR | OK)
BOB_status    = ((ERROR:(result, info)                   -> (result, info, {}))
                |(OK_DATA_TUNNEL_STATUS:(result, status) -> (result, '', status)))
BOB_stop      = (ERROR | OK)
BOB_verify    = (ERROR | OK)
BOB_visit     = (OK)

State_init      = BOB_init:version      -> receiver.initBOB(version)

State_clear     = BOB_clear:response     -> receiver.clear(*response)
State_getdest   = BOB_getdest:response   -> receiver.getdest(*response)
State_getkeys   = BOB_getkeys:response   -> receiver.getkeys(*response)
State_getnick   = BOB_getnick:response   -> receiver.getnick(*response)
State_inhost    = BOB_inhost:response    -> receiver.inhost(*response)
State_inport    = BOB_inport:response    -> receiver.inport(*response)
State_list      = BOB_list:response      -> receiver.list(*response)
State_newkeys   = BOB_newkeys:response   -> receiver.newkeys(*response)
State_option    = BOB_option:response    -> receiver.option(*response)
State_outhost   = BOB_outhost:response   -> receiver.outhost(*response)
State_outport   = BOB_outport:response   -> receiver.outport(*response)
State_quiet     = BOB_quiet:response     -> receiver.quiet(*response)
State_quit      = BOB_quit:response      -> receiver.quit(*response)
State_setkeys   = BOB_setkeys:response   -> receiver.setkeys(*response)
State_setnick   = BOB_setnick:response   -> receiver.setnick(*response)
State_show      = BOB_show:response      -> receiver.show(*response)
State_showprops = BOB_showprops:response -> receiver.showprops(*response)
State_start     = BOB_start:response     -> receiver.start(*response)
State_status    = BOB_status:response    -> receiver.status(*response)
State_stop      = BOB_stop:response      -> receiver.stop(*response)
State_verify    = BOB_verify:response    -> receiver.verify(*response)
State_visit     = BOB_visit:response     -> receiver.visit(*response)
"""

# SAM grammar
samGrammarSource = i2pGrammarSource + r"""
KEY = <(~'=' anything)*>
VALUE = (('"' <(~'"' anything)*>:value '"' -> value)
        |(<(~(' '|'\n') anything)*>))
OPTION = KEY:key '=' VALUE:value -> (key.lower(), value)
OPTIONS = (OPTION:first (' ' OPTION)*:rest -> dict([first] + rest))
          | -> {}

SAM_hello          = 'HELLO REPLY '    OPTIONS:options '\n' -> options
SAM_session_status = 'SESSION STATUS ' OPTIONS:options '\n' -> options
SAM_stream_status  = 'STREAM STATUS '  OPTIONS:options '\n' -> options
SAM_naming_reply   = 'NAMING REPLY '   OPTIONS:options '\n' -> options
SAM_dest_reply     = 'DEST REPLY '     OPTIONS:options '\n' -> options

State_hello   = SAM_hello:options          -> receiver.hello(**options)
State_create  = SAM_session_status:options -> receiver.create(**options)
State_connect = SAM_stream_status:options  -> receiver.connect(**options)
State_accept  = SAM_stream_status:options  -> receiver.accept(**options)
State_forward = SAM_stream_status:options  -> receiver.forward(**options)
State_naming  = SAM_naming_reply:options   -> receiver.lookupReply(**options)
State_dest    = SAM_dest_reply:options     -> receiver.destGenerated(**options)
State_readData = anything:data             -> receiver.dataReceived(data)

KEEPALIVE_DATA = (' '+ <(~'\n' anything)+>:data -> data) | -> None
SAM_ping = 'PING' KEEPALIVE_DATA:data '\n' -> data
SAM_pong = 'PONG' KEEPALIVE_DATA:data '\n' -> data

State_keepalive = ((SAM_ping:data -> receiver.ping(data))
                  |(SAM_pong:data -> receiver.pong(data)))
"""



try:
    from txi2p._version import __version__, __version_tuple__
except ImportError:
    __version__ = __version_tuple__ = None

from txi2p.address import I2PAddress
from txi2p.utils import generateDestination, testAPI



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
from twisted.internet import defer
from twisted.internet.error import ConnectionLost, ConnectionRefusedError
from twisted.python import failure
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.bob.factory import (BOBI2PClientFactory,
                               BOBI2PServerFactory,
                               BOBClientFactoryWrapper,
                               BOBServerFactoryWrapper)
from txi2p.test.util import FakeFactory

connectionLostFailure = failure.Failure(ConnectionLost())
connectionRefusedFailure = failure.Failure(ConnectionRefusedError())


class BOBFactoryTestMixin(object):
    def setUp(self):
        self.aborted = []

    def makeProto(self, *a, **kw):
        fac = self.factory(*a, **kw)
        proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: self.aborted.append(True)
        proto.makeConnection(transport)
        return fac, proto

    def test_cancellation(self):
        fac, proto = self.makeProto(None, None, None, '')
        fac.deferred.cancel()
        self.assert_(self.aborted)
        return self.assertFailure(fac.deferred, defer.CancelledError)

    def test_cancellationBeforeFailure(self):
        fac, proto = self.makeProto(None, None, None, '')
        fac.deferred.cancel()
        proto.connectionLost(connectionLostFailure)
        self.assert_(self.aborted)
        return self.assertFailure(fac.deferred, defer.CancelledError)

    def test_cancellationAfterFailure(self):
        fac, proto = self.makeProto(None, None, None, '')
        proto.connectionLost(connectionLostFailure)
        fac.deferred.cancel()
        self.assertFalse(self.aborted)
        return self.assertFailure(fac.deferred, ConnectionLost)

    def test_clientConnectionFailed(self):
        fac, proto = self.makeProto(None, None, None, '')
        fac.clientConnectionFailed(None, connectionRefusedFailure)
        return self.assertFailure(fac.deferred, ConnectionRefusedError)

    def test_defaultFactoryListsTunnels(self):
        fac, proto = self.makeProto(None, None, None, '')
        proto.dataReceived('BOB 00.00.10\nOK\n')
        self.assertEqual(proto.transport.value(), b'list\n')


class TestBOBI2PClientFactory(BOBFactoryTestMixin, unittest.TestCase):
    factory = BOBI2PClientFactory

    def TODO_test_noProtocolFromWrappedFactory(self):
        wrappedFac = FakeFactory(returnNoProtocol=True)
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto(mreactor, wrappedFac, None, '')
        fac.tunnelNick = 'spam'
        # Shortcut to end of BOB protocol
        proto.receiver.currentRule = 'State_start'
        proto._parser._setupInterp()
        proto.dataReceived('OK HTTP 418\n')
        self.assert_(self.aborted) # TODO: Check the Deferred chain
        return self.assertFailure(fac.deferred, defer.CancelledError)


class TestBOBI2PServerFactory(BOBFactoryTestMixin, unittest.TestCase):
    factory = BOBI2PServerFactory

    def TODO_test_noProtocolFromWrappedFactory(self):
        wrappedFac = FakeFactory(returnNoProtocol=True)
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto(mreactor, wrappedFac, None, '')
        fac.tunnelNick = 'spam'
        fac.localDest = 'spam.i2p'
        # Shortcut to end of BOB protocol
        proto.receiver.currentRule = 'State_start'
        proto._parser._setupInterp()
        proto.dataReceived('OK HTTP 418\n')
        self.assert_(self.aborted) # TODO: Check the Deferred chain
        return self.assertFailure(fac.deferred, defer.CancelledError)


class TestBOBClientFactoryWrapper(unittest.TestCase):
    def test_buildProtocol(self):
        wrappedFac = FakeFactory()
        fac = BOBClientFactoryWrapper(wrappedFac, None, None, '', True)
        fac.setDest('spam.i2p')
        proto = fac.buildProtocol(None)
        self.assertEqual(proto.wrappedProto.factory, wrappedFac)


class TestBOBServerFactoryWrapper(unittest.TestCase):
    def test_buildProtocol(self):
        wrappedFac = FakeFactory()
        fac = BOBServerFactoryWrapper(wrappedFac, None, None, '', True)
        proto = fac.buildProtocol(None)
        self.assertEqual(proto.wrappedProto.factory, wrappedFac)



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
try:
    # Python 3
    from unittest import mock
except:
    # Python 2 (library)
    import mock
from twisted.internet import defer, protocol
from twisted.test import proto_helpers

TEST_B64 = "2wDRF5nDfeTNgM4X-TI5xEk3R-WiaTABvkMQ2eYpvEzayUZQJgr9E2T6Y2m9HHn3xHYGEOg-RLisjW9AubTaUTx-v66AsEEtv745qPcuWuV1SP~w1bdzYEn8MSoK7Zh4mwHBg1uHq8z17TUNvWz19q76vHNth-2PDuBToD7ySBn3cGBFDUU83wJJXPD6OueLY8yosWWtksk7WZk60~6z~nVePPSEY8JDry3myLDe11szAVER4A8eX1sFpw247cXGGJK9wQhV-TXFj~m76GPVcFKh7u79zwTwZnZ1GXXKqqyRoj1c4-U69CvvJsQRLmdLFwFEpRkxwV8z6LIFclYJk443YpTnPXC7vNdFOzqqS4FLR1ra~DNfN5foMtR2~2VxuR5m2dYiOS6GzHDxA4acJJSGqnasJjcEIFNVSQKxMnFu9PvGLNJHZ83EraHCErENcOGkPlnVgcJCtPGNGiirwCbBz38jE0lfjkrNrWabc6uWeU559CobG8F8KUDx1irpAAAA"
TEST_B32 = "tv5iv4i5roywnv2rg6rjqufniqbogn4rokjkooa7n4jht3lex3ga.b32.i2p"


def fakeSession(nickname, **kwargs):
    m = mock.Mock()
    m.nickname = nickname
    m.kwargs = kwargs
    return m


class FakeFactory(protocol.ClientFactory):
    protocol = proto_helpers.AccumulatingProtocol

    def __init__(self, returnNoProtocol=False):
        self.returnNoProtocol = returnNoProtocol
        self.protocolConnectionMade = defer.Deferred()

    def buildProtocol(self, addr):
        if self.returnNoProtocol:
            return None
        self.proto = protocol.ClientFactory.buildProtocol(self, addr)
        return self.proto


class FakeEndpoint(object):
    def __init__(self, failure=None):
        self.failure = failure
        self.deferred = None

    def connect(self, fac):
        self.factory = fac
        if self.deferred:
            return self.deferred
        if self.failure:
            return defer.fail(self.failure)
        self.proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        self.aborted = []
        transport.abortConnection = lambda: self.aborted.append(True)
        self.tlsStarts = []
        transport.startTLS = lambda ctx: self.tlsStarts.append(ctx)
        self.proto.makeConnection(transport)
        self.transport = transport
        return defer.succeed(self.proto)



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.endpoints import clientFromString
from twisted.internet.interfaces import IStreamClientEndpointStringParserWithReactor
from twisted.internet.interfaces import IStreamServerEndpointStringParser
from zope.interface import implementer

from txi2p.bob.endpoints import BOBI2PClientEndpoint, BOBI2PServerEndpoint
from txi2p.sam.endpoints import (
    SAMI2PStreamClientEndpoint,
    SAMI2PStreamServerEndpoint,
)
from txi2p.utils import getApi

from twisted.plugin import IPlugin


def _parseOptions(options):
    return dict([option.split(':') for option in options.split(',')]) if options else {}

@implementer(IPlugin, IStreamClientEndpointStringParserWithReactor)
class I2PClientParser:
    prefix = 'i2p'

    def _parseBOBClient(self, reactor, host, port, bobEndpoint,
                     tunnelNick=None,
                     inhost='localhost',
                     inport=None,
                     options=None):
        return BOBI2PClientEndpoint(reactor, clientFromString(reactor, bobEndpoint),
                                    host, port, tunnelNick, inhost,
                                    inport and int(inport) or None,
                                    _parseOptions(options))

    def _parseSAMClient(self, reactor, host, port, samEndpoint,
                     nickname=None,
                     autoClose=False,
                     keyfile=None,
                     localPort=None,
                     options=None,
                     sigType=None):
        return SAMI2PStreamClientEndpoint.new(
            clientFromString(reactor, samEndpoint),
            host, port, nickname, autoClose, keyfile,
            localPort and int(localPort) or None, _parseOptions(options), sigType)

    _apiParsers = {
        'BOB': _parseBOBClient,
        'SAM': _parseSAMClient,
        }

    def _parseClient(self, reactor, host, port=None,
                     api=None, apiEndpoint=None, **kwargs):
        api, apiEndpoint = getApi(api, apiEndpoint, self._apiParsers)
        return self._apiParsers[api](self, reactor, host,
                                     port and int(port) or None,
                                     apiEndpoint, **kwargs)

    def parseStreamClient(self, reactor, *args, **kwargs):
        # Delegate to another function with a sane signature.  This function has
        # an insane signature to trick zope.interface into believing the
        # interface is correctly implemented.
        return self._parseClient(reactor, *args, **kwargs)


@implementer(IPlugin, IStreamServerEndpointStringParser)
class I2PServerParser:
    prefix = 'i2p'

    def _parseBOBServer(self, reactor, keyfile, port, bobEndpoint,
                     tunnelNick=None,
                     outhost='localhost',
                     outport=None,
                     options=None):
        return BOBI2PServerEndpoint(reactor, clientFromString(reactor, bobEndpoint),
                                    keyfile, port, tunnelNick, outhost,
                                    outport and int(outport) or None,
                                    _parseOptions(options))

    def _parseSAMServer(self, reactor, keyfile, port, samEndpoint,
                     nickname=None,
                     autoClose=False,
                     options=None,
                     sigType=None):
        return SAMI2PStreamServerEndpoint.new(
            clientFromString(reactor, samEndpoint),
            keyfile, port, nickname, autoClose, _parseOptions(options), sigType)

    _apiParsers = {
        'BOB': _parseBOBServer,
        'SAM': _parseSAMServer,
        }

    def _parseServer(self, reactor, keyfile, port=None,
                     api=None, apiEndpoint=None, **kwargs):
        api, apiEndpoint = getApi(api, apiEndpoint, self._apiParsers)
        return self._apiParsers[api](self, reactor, keyfile,
                                     port and int(port) or None,
                                     apiEndpoint, **kwargs)

    def parseStreamServer(self, reactor, *args, **kwargs):
        # Delegate to another function with a sane signature.  This function has
        # an insane signature to trick zope.interface into believing the
        # interface is correctly implemented.
        return self._parseServer(reactor, *args, **kwargs)



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
try:
    # Python 3
    from unittest import mock
except:
    # Python 2 (library)
    import mock
import sys
from twisted.internet import interfaces
from twisted.python.versions import Version
from twisted.test.proto_helpers import MemoryReactor
from twisted.trial import unittest
import twisted
from zope.interface.verify import verifyObject

from txi2p.bob.endpoints import (BOBI2PClientEndpoint,
                                 BOBI2PServerEndpoint)
from txi2p.sam.endpoints import (SAMI2PStreamClientEndpoint,
                                 SAMI2PStreamServerEndpoint)
from txi2p.test.util import fakeSession

if twisted.version < Version('twisted', 14, 0, 0):
    skip = 'txi2p.plugins requires twisted 14.0 or newer'
else:
    skip = None


class I2PPluginTestMixin(object):
    def test_pluginDiscovery(self):
        from twisted.internet.endpoints import getPlugins
        parsers = list(getPlugins(self._parserInterface))
        for p in parsers:
            if isinstance(p, self._parserClass):
                break
        else:
            self.fail(
                "Did not find %s parser in %r" % (self._parserClass, parsers,))

    def test_interface(self):
        parser = self._parserClass()
        self.assertTrue(verifyObject(self._parserInterface, parser))


class I2PClientEndpointPluginTest(I2PPluginTestMixin, unittest.TestCase):
    """
    Unit tests for the I2P client endpoint description parser.
    """

    skip = skip
    @property
    def _parserInterface(self):
        return interfaces.IStreamClientEndpointStringParserWithReactor
    @property
    def _parserClass(self):
        from txi2p.plugins import I2PClientParser
        return I2PClientParser

    def test_badAPI(self):
        from twisted.internet.endpoints import clientFromString
        self.failUnlessRaises(ValueError, clientFromString,
            MemoryReactor(), "i2p:stats.i2p:api=FOO")

    def test_apiEndpointWithNoAPI(self):
        from twisted.internet.endpoints import clientFromString
        self.failUnlessRaises(ValueError, clientFromString,
            MemoryReactor(), "i2p:stats.i2p:apiEndpoint=tcp\:127.0.0.1\:2827")

    def test_stringDescription_default(self):
        from twisted.internet.endpoints import clientFromString
        with mock.patch('txi2p.sam.endpoints.getSession', fakeSession):
            ep = clientFromString(
                MemoryReactor(), "i2p:stats.i2p")
        self.assertIsInstance(ep, SAMI2PStreamClientEndpoint)

    def test_stringDescription_BOB(self):
        from twisted.internet.endpoints import clientFromString
        ep = clientFromString(
            MemoryReactor(), "i2p:stats.i2p:api=BOB:tunnelNick=spam:inport=12345:options=inbound.length\:5,outbound.length\:5")
        self.assertIsInstance(ep, BOBI2PClientEndpoint)
        self.assertIsInstance(ep._reactor, MemoryReactor)
        self.assertEqual(ep._dest,"stats.i2p")
        self.assertEqual(ep._tunnelNick,"spam")
        self.assertEqual(ep._inport,12345)
        self.assertEqual(ep._options, {'inbound.length': '5', 'outbound.length': '5'})

    def test_stringDescription_SAM(self):
        from twisted.internet.endpoints import clientFromString
        with mock.patch('txi2p.sam.endpoints.getSession', fakeSession):
            ep = clientFromString(
                MemoryReactor(), "i2p:stats.i2p:81:api=SAM:localPort=34444:options=inbound.length\:5,outbound.length\:5:sigType=foobar")
        self.assertIsInstance(ep, SAMI2PStreamClientEndpoint)
        self.assertEqual(ep._host, "stats.i2p")
        self.assertEqual(ep._port, 81)
        self.assertEqual(ep._localPort, 34444)
        s = ep._sessionDeferred
        self.assertEqual(s.kwargs['options'], {'inbound.length': '5', 'outbound.length': '5'})
        self.assertEqual(s.kwargs['sigType'], 'foobar')


class I2PServerEndpointPluginTest(I2PPluginTestMixin, unittest.TestCase):
    """
    Unit tests for the I2P client endpoint description parser.
    """

    skip = skip
    @property
    def _parserInterface(self):
        return interfaces.IStreamServerEndpointStringParser
    @property
    def _parserClass(self):
        from txi2p.plugins import I2PServerParser
        return I2PServerParser

    def test_badAPI(self):
        from twisted.internet.endpoints import serverFromString
        self.failUnlessRaises(ValueError, serverFromString,
            MemoryReactor(), "i2p:/tmp/testkeys.foo:api=FOO")

    def test_apiEndpointWithNoAPI(self):
        from twisted.internet.endpoints import serverFromString
        self.failUnlessRaises(ValueError, serverFromString,
            MemoryReactor(), "i2p:/tmp/testkeys.foo:apiEndpoint=tcp\:127.0.0.1\:2827")

    def test_stringDescription_default(self):
        from twisted.internet.endpoints import serverFromString
        with mock.patch('txi2p.sam.endpoints.getSession', fakeSession):
            ep = serverFromString(
                MemoryReactor(), "i2p:/tmp/testkeys.foo")
        self.assertIsInstance(ep, SAMI2PStreamServerEndpoint)

    def test_stringDescription_BOB(self):
        from twisted.internet.endpoints import serverFromString
        ep = serverFromString(
            MemoryReactor(), "i2p:/tmp/testkeys.foo:api=BOB:tunnelNick=spam:outport=23456:options=inbound.length\:5,outbound.length\:5")
        self.assertIsInstance(ep, BOBI2PServerEndpoint)
        self.assertIsInstance(ep._reactor, MemoryReactor)
        self.assertEqual(ep._keyfile, "/tmp/testkeys.foo")
        self.assertEqual(ep._tunnelNick, "spam")
        self.assertEqual(ep._outport, 23456)
        self.assertEqual(ep._options, {'inbound.length': '5', 'outbound.length': '5'})

    def test_stringDescription_SAM(self):
        from twisted.internet.endpoints import serverFromString
        with mock.patch('txi2p.sam.endpoints.getSession', fakeSession):
            ep = serverFromString(
                MemoryReactor(), "i2p:/tmp/testkeys.foo:81:api=SAM:options=inbound.length\:5,outbound.length\:5:sigType=foobar")
        self.assertIsInstance(ep, SAMI2PStreamServerEndpoint)
        s = ep._sessionDeferred
        self.assertEqual(s.kwargs['options'], {'inbound.length': '5', 'outbound.length': '5'})
        self.assertEqual(s.kwargs['sigType'], 'foobar')



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.error import ConnectionLost, ConnectionRefusedError
from twisted.internet.protocol import Factory, Protocol
from twisted.internet.interfaces import IStreamServerEndpoint
from twisted.python import failure
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.sam import endpoints
from txi2p.sam.session import SAMSession
from txi2p.test.util import FakeEndpoint, FakeFactory, fakeSession


connectionLostFailure = failure.Failure(ConnectionLost())
connectionRefusedFailure = failure.Failure(ConnectionRefusedError())


def sessionOptionsFromNew(cls, **kwargs):
    """
    Create a new session factory and return the value of its ``options``
    attribute.

    :param cls: A class like ``endpoints.SAMI2PStreamClientEndpoint``.

    :return: A ``Deferred`` that fires with the ``options`` value.
    """
    samEndpoint = FakeEndpoint(failure=connectionRefusedFailure)
    endpoint = cls.new(samEndpoint, '', **kwargs)
    if IStreamServerEndpoint.providedBy(endpoint):
        f = endpoint.listen
    else:
        f = endpoint.connect
    d = f(Factory.forProtocol(Protocol))
    d.addErrback(lambda err: err.trap(ConnectionRefusedError))
    d.addCallback(lambda ignored: samEndpoint.factory.options)
    return d


class SAMI2PStreamClientEndpointTestCase(unittest.TestCase):
    """
    Tests for I2P client Endpoint backed by the SAM API.
    """
    endpointClass = endpoints.SAMI2PStreamClientEndpoint

    def test_newWithOptions(self):
        """
        If ``SAMI2PStreamClientEndpoint.new`` is called options then
        ``SessionCreateFactory`` is created with those options.
        """
        options = {'inbound.length': 5, 'outbound.length': 5}
        d = sessionOptionsFromNew(self.endpointClass, options=options)
        self.assertEqual(self.successResultOf(d), options)

    def test_newWithoutOptions(self):
        """
        If ``SAMI2PStreamClientEndpoint.new`` is called without options then
        ``SessionCreateFactory`` is created with an empty options dictionary.
        """
        d = sessionOptionsFromNew(self.endpointClass)
        self.assertEqual(self.successResultOf(d), {})

    def test_samConnectionFailed(self):
        samEndpoint = FakeEndpoint(failure=connectionRefusedFailure)
        endpoint = endpoints.SAMI2PStreamClientEndpoint.new(samEndpoint, '')
        d = endpoint.connect(Factory.forProtocol(Protocol))
        self.failureResultOf(d, ConnectionRefusedError)

    def test_streamConnect(self):
        samEndpoint = FakeEndpoint()
        session = SAMSession()
        session.nickname = 'foo'
        session.samEndpoint = samEndpoint
        session.samVersion = '3.1'
        session.id = 'foo'
        session._autoClose = True
        endpoint = endpoints.SAMI2PStreamClientEndpoint(session, 'foo.i2p')
        endpoint.connect(None)
        self.assertSubstring('HELLO VERSION', samEndpoint.transport.value().decode('utf-8'))



class SAMI2PStreamServerEndpointTestCase(unittest.TestCase):
    """
    Tests for I2P server Endpoint backed by the SAM API.
    """
    endpointClass = endpoints.SAMI2PStreamServerEndpoint

    def test_newWithOptions(self):
        """
        If ``SAMI2PStreamServerEndpoint.new`` is called options then
        ``SessionCreateFactory`` is created with those options.
        """
        options = {'inbound.length': 5, 'outbound.length': 5}
        d = sessionOptionsFromNew(self.endpointClass, options=options)
        self.assertEqual(self.successResultOf(d), options)

    def test_newWithoutOptions(self):
        """
        If ``SAMI2PStreamServerEndpoint.new`` is called without options then
        ``SessionCreateFactory`` is created with an empty options dictionary.
        """
        d = sessionOptionsFromNew(self.endpointClass)
        self.assertEqual(self.successResultOf(d), {})

    def test_samConnectionFailed(self):
        samEndpoint = FakeEndpoint(failure=connectionRefusedFailure)
        endpoint = endpoints.SAMI2PStreamServerEndpoint.new(samEndpoint, '')
        d = endpoint.listen(Factory.forProtocol(Protocol))
        self.failureResultOf(d, ConnectionRefusedError)

    def test_streamListen(self):
        samEndpoint = FakeEndpoint()
        session = SAMSession()
        session.nickname = 'foo'
        session.samEndpoint = samEndpoint
        session.samVersion = '3.1'
        session.id = 'foo'
        session._autoClose = True
        endpoint = endpoints.SAMI2PStreamServerEndpoint(session)
        endpoint.listen(None)
        self.assertSubstring('HELLO VERSION', str(samEndpoint.transport.value()))



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
from twisted.internet import defer, error, interfaces
from twisted.internet.endpoints import serverFromString
from zope.interface import implementer

from txi2p.sam.base import I2PFactoryWrapper
from txi2p.sam.session import SAMSession, getSession
from txi2p.sam.stream import (
    StreamConnectFactory,
    StreamAcceptPort,
    StreamForwardFactory,
    StreamForwardPort,
)


def _parseHost(host):
    # TODO: Validate I2P domain, B32 etc.
    return (host, None) if host[-4:] == '.i2p' else (None, host)


@implementer(interfaces.IStreamClientEndpoint)
class SAMI2PStreamClientEndpoint(object):
    """I2P stream client endpoint backed by the SAM API.

    Args:
        session (txi2p.sam.SAMSession): The SAM session to connect with.
        host (str): The I2P hostname or Destination to connect to.
        port (int): The port to connect to inside I2P. If unset or `None`, the
            default (null) port is used. Ignored if the SAM server doesn't
            support SAM v3.2 or higher.
        localPort (int): The port to connect from inside I2P. This can be used
            to distinguish between multiple connections to the same server. If
            unset or `None`, the default (null) port is used. Ignored if the SAM
            server doesn't support SAM v3.2 or higher.
    """

    @classmethod
    def new(cls, samEndpoint, host, port=None, nickname=None, autoClose=False, keyfile=None, localPort=None, options=None, sigType=None):
        """Create an I2P client endpoint backed by the SAM API.

        If a SAM session for ``nickname`` already exists, it will be used, and
        all options other than ``host`` and ``port`` will be ignored. Otherwise,
        a new SAM session will be created. The implication of this is that by
        default, all endpoints (both client and server) created by the same
        process will use the same SAM session.

        Args:
            samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
                endpoint that will connect to the SAM API.
            host (str): The I2P hostname or Destination to connect to.
            port (int): The port to connect to inside I2P. If unset or `None`,
                the default (null) port is used. Ignored if the SAM server
                doesn't support SAM v3.2 or higher.
            nickname (str): The SAM session nickname.
            autoClose (bool): `true` if the session should close automatically
                once no more connections are using it.
            keyfile (str): Path to a local file containing the keypair to use
                for the session Destination. If non-existent, new keys will be
                generated and stored.
            localPort (int): The port to connect from inside I2P. This can be
                used to distinguish between multiple connections to the same
                server. If unset or `None`, the default (null) port is used.
                Ignored if the SAM server doesn't support SAM v3.2 or higher.
            options (dict): I2CP options to configure the session with.
            sigType (str): The SigType to use if generating a new Destination.
                Defaults to Ed25519 if supported, falling back to
                ECDSA_SHA256_P256 and then DSA_SHA1.
        """
        d = getSession(nickname,
                       samEndpoint=samEndpoint,
                       autoClose=autoClose,
                       keyfile=keyfile,
                       options=options,
                       sigType=sigType)
        return cls(d, host, port, localPort)

    def __init__(self, session, host, port=None, localPort=None):
        self._host, self._dest = _parseHost(host)
        self._port = port
        self._localPort = localPort
        if isinstance(session, SAMSession):
            self._session = session
        else:
            self._session = None
            self._sessionDeferred = session

    def connect(self, fac):
        """Connect over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P client tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        def createStream(val):
            if self._session.style != 'STREAM':
                raise error.UnsupportedSocketType()

            i2pFac = StreamConnectFactory(fac, self._session, self._host, self._dest, self._port, self._localPort)
            d = self._session.samEndpoint.connect(i2pFac)
            # Once the SAM IProtocol is returned, wait for the
            # real IProtocol to be returned after tunnel creation,
            # and pass it to any further registered callbacks.
            d.addCallback(lambda proto: i2pFac.deferred)
            return d

        if self._session:
            return createStream(None)

        def saveSession(session):
            self._session = session
            return None
        self._sessionDeferred.addCallback(saveSession)
        self._sessionDeferred.addCallback(createStream)
        return self._sessionDeferred


@implementer(interfaces.IStreamServerEndpoint)
class SAMI2PStreamServerEndpoint(object):
    """I2P server endpoint backed by the SAM API.

    Args:
        session (txi2p.sam.SAMSession): The SAM session to listen on.
    """

    @classmethod
    def new(cls, samEndpoint, keyfile, port=None, nickname=None, autoClose=False, options=None, sigType=None):
        """Create an I2P server endpoint backed by the SAM API.

        If a SAM session for ``nickname`` already exists, it will be used, and
        all options other than ``port`` will be ignored. Otherwise, a new SAM
        session will be created. The implication of this is that by default, all
        endpoints (both client and server) created by the same process will use
        the same SAM session.

        Args:
            samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
                endpoint that will connect to the SAM API.
            keyfile (str): Path to a local file containing the keypair to use
                for the session Destination. If non-existent, new keys will be
                generated and stored.
            port (int): The port to listen on inside I2P. If unset or `None`,
                the default (null) port is used. Ignored if the SAM server
                doesn't support SAM v3.2 or higher.
            nickname (str): The SAM session nickname.
            autoClose (bool): `true` if the session should close automatically
                once no more connections are using it.
            options (dict): I2CP options to configure the session with.
            sigType (str): The SigType to use if generating a new Destination.
                Defaults to Ed25519 if supported, falling back to
                ECDSA_SHA256_P256 and then DSA_SHA1.
        """
        d = getSession(nickname,
                       samEndpoint=samEndpoint,
                       autoClose=autoClose,
                       keyfile=keyfile,
                       localPort=port,
                       options=options,
                       sigType=sigType)
        return cls(d)

    def __init__(self, session):
        if isinstance(session, SAMSession):
            self._session = session
        else:
            self._session = None
            self._sessionDeferred = session

    def listen(self, fac):
        """Listen over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P server tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        def createAcceptingStream(val):
            if self._session.style != 'STREAM':
                raise error.UnsupportedSocketType()

            p = StreamAcceptPort(self._session, fac)
            p.startListening()
            return p

#        def createForwardingStream(val):
#            if self._session.style != 'STREAM':
#                raise error.UnsupportedSocketType()
#
#            serverEndpoint = serverFromString(self._reactor,
#                                              'tcp:0:interface=127.0.0.1')
#            wrappedFactory = I2PFactoryWrapper(fac, self._session.address)
#            d = serverEndpoint.listen(wrappedFactory)
#
#            def setupForward(port):
#                local_port = port.getHost().port
#                i2pFac = StreamForwardFactory(self._session, local_port)
#                d2 = self._session.samEndpoint.connect(i2pFac)
#                d2.addCallback(lambda proto: i2pFac.deferred)
#                d2.addCallback(lambda forwardingProto: (port, forwardingProto))
#                return d2
#
#            def handlePort((port, forwardingProto)):
#                return StreamForwardPort(port, forwardingProto, self._session.address)
#
#            d.addCallback(setupForward)
#            d.addCallback(handlePort)
#            return d

        if self._session:
            return createAcceptingStream(None)

        def saveSession(session):
            self._session = session
            return None
        self._sessionDeferred.addCallback(saveSession)
        self._sessionDeferred.addCallback(createAcceptingStream)
        return self._sessionDeferred



from __future__ import print_function
from twisted.internet import reactor, defer
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver


class Answer(LineReceiver):
    answers = {
        'How are you?': 'Fine',
        'W007!': 'INORITE?',
        None : "I don't know what you mean",
        }

    def lineReceived(self, line):
        print('Line received from %s' % self.transport.getPeer())
        if self.answers.has_key(line):
            self.sendLine(self.answers[line])
        else:
            self.sendLine(self.answers[None])


class AnswerFactory(Factory):
    protocol = Answer


def printDest(port):
    # Print out the I2P Destination to copy to the client
    print('This server is listening on:')
    print(port.getHost().destination)
    # Handle Ctl+C
    def shutdown():
        print('Shutting down')
        port.stopListening()
        d = defer.Deferred()
        reactor.callLater(3, d.callback, 1)
        return d
    reactor.addSystemEventTrigger('before', 'shutdown', shutdown)


endpoint = serverFromString(reactor, 'i2p:keypair.answerserver')
d = endpoint.listen(AnswerFactory())
d.addCallback(printDest)

reactor.run()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
from twisted.internet import interfaces
from zope.interface import implementer

from txi2p.bob.factory import BOBI2PClientFactory, BOBI2PServerFactory


def _validateDestination(dest):
    # TODO: Validate I2P domain, B32 etc.
    pass


@implementer(interfaces.IStreamClientEndpoint)
class BOBI2PClientEndpoint(object):
    """I2P client endpoint backed by the BOB API.

    Args:
        reactor: The client endpoint will be constructed with this reactor.
        bobEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that will connect to the BOB API.
        host (str): The I2P hostname or Destination to connect to.
        port (int): The port to connect to inside I2P. If unset or `None`, the
            default (null) port is used. Ignored because BOB doesn't support
            ports yet.
        tunnelNick (str): The tunnel nickname to use. If a tunnel with this
            nickname already exists, it will be used. The default is ``txi2p-#``
            where ``#`` is the PID of the current process.

            * The implication of this is that by default, all endpoints (both
              client and server) created by the same process will use the same
              BOB tunnel.

        inhost (str): The host that the tunnel created by BOB will listen on.
            Defaults to ``localhost``.
        inport (int): The port that the tunnel created by BOB will listen on.
            Defaults to a port over 9000.
        options (dict): I2CP options to configure the tunnel with.
    """

    def __init__(self, reactor, bobEndpoint, dest,
                 port=None,
                 tunnelNick=None,
                 inhost='localhost',
                 inport=None,
                 options=None):
        _validateDestination(dest)
        self._reactor = reactor
        self._bobEndpoint = bobEndpoint
        self._dest = dest
        self._port = port
        self._tunnelNick = tunnelNick
        self._inhost = inhost
        self._inport = inport
        self._options = options

    def connect(self, fac):
        """Connect over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P client tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        i2pFac = BOBI2PClientFactory(self._reactor, fac, self._bobEndpoint, self._dest,
                                     self._tunnelNick,
                                     self._inhost,
                                     self._inport,
                                     self._options)
        d = self._bobEndpoint.connect(i2pFac)
        # Once the BOB IProtocol is returned, wait for the
        # real IProtocol to be returned after tunnel creation,
        # and pass it to any further registered callbacks.
        d.addCallback(lambda proto: i2pFac.deferred)
        return d


@implementer(interfaces.IStreamServerEndpoint)
class BOBI2PServerEndpoint(object):
    """I2P server endpoint backed by the BOB API.

    Args:
        reactor: The server endpoint will be constructed with this reactor.
        bobEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that will connect to the BOB API.
        keyfile (str): Path to a local file containing the keypair to use for
            the server Destination. If non-existent, new keys will be generated
            and stored.
        port (int): The port to connect to inside I2P. If unset or `None`, the
            default (null) port is used. Ignored because BOB doesn't support
            ports yet.
        tunnelNick (str): The tunnel nickname to use. If a tunnel with this
            nickname already exists, it will be used. The default is ``txi2p-#``
            where ``#`` is the PID of the current process.

            * The implication of this is that by default, all endpoints (both
              client and server) created by the same process will use the same
              BOB tunnel.

        outhost (str): The host that the tunnel created by BOB will forward data
            to. Defaults to ``localhost``.
        outport (int): The port that the tunnel created by BOB will forward data
            to. Defaults to a port over 9000.
        options (dict): I2CP options to configure the tunnel with.
    """

    def __init__(self, reactor, bobEndpoint, keyfile,
                 port=None,
                 tunnelNick=None,
                 outhost='localhost',
                 outport=None,
                 options=None):
        self._reactor = reactor
        self._bobEndpoint = bobEndpoint
        self._keyfile = keyfile
        self._port = port
        self._tunnelNick = tunnelNick
        self._outhost = outhost
        self._outport = outport
        self._options = options

    def listen(self, fac):
        """Listen over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P server tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        i2pFac = BOBI2PServerFactory(self._reactor, fac, self._bobEndpoint, self._keyfile,
                                     self._tunnelNick,
                                     self._outhost,
                                     self._outport,
                                     self._options)
        d = self._bobEndpoint.connect(i2pFac)
        # Once the BOB IProtocol is returned, wait for the
        # IListeningPort to be returned after tunnel creation,
        # and pass it to any further registered callbacks.
        d.addCallback(lambda proto: i2pFac.deferred)
        return d



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

import unittest

from txi2p.address import I2PAddress
from txi2p.test.util import TEST_B64, TEST_B32


class TestI2PAddress(unittest.TestCase):
    def test_noHost_noPort(self):
        addr = I2PAddress(TEST_B64)
        self.assertEqual((addr.destination, addr.host, addr.port), (TEST_B64, TEST_B32, None))

    def test_host_noPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p')
        self.assertEqual((addr.destination, addr.host, addr.port), (TEST_B64, 'spam.i2p', None))

    def test_noHost_port(self):
        addr = I2PAddress(TEST_B64, port=81)
        self.assertEqual((addr.destination, addr.host, addr.port), (TEST_B64, TEST_B32, 81))

    def test_host_port(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        self.assertEqual((addr.destination, addr.host, addr.port), (TEST_B64, 'spam.i2p', 81))

    def test_portString(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', '81')
        self.assertEqual((addr.destination, addr.host, addr.port), (TEST_B64, 'spam.i2p', 81))

    def test_address_host_noPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        addr2 = I2PAddress(addr, host='eggs.i2p')
        self.assertEqual((addr2.destination, addr2.host, addr2.port), (TEST_B64, 'eggs.i2p', None))

    def test_address_noHost_port(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        addr2 = I2PAddress(addr, port=82)
        self.assertEqual((addr2.destination, addr2.host, addr2.port), (TEST_B64, 'spam.i2p', 82))

    def test_address_host_port(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        addr2 = I2PAddress(addr, host='eggs.i2p', port=82)
        self.assertEqual((addr2.destination, addr2.host, addr2.port), (TEST_B64, 'eggs.i2p', 82))

    def test_reprWithNoHostNoPort(self):
        addr = I2PAddress(TEST_B64)
        self.assertEqual(repr(addr), 'I2PAddress(%s)' % TEST_B32)

    def test_reprWithHostNoPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p')
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p)')

    def test_reprWithNoHostPort(self):
        addr = I2PAddress(TEST_B64, port=81)
        self.assertEqual(repr(addr), 'I2PAddress(%s, 81)' % TEST_B32)

    def test_reprWithHostPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p, 81)')

    def test_reprWithPortString(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', '81')
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p, 81)')

    def test_hashWithNoHostNoPort(self):
        addr = I2PAddress(TEST_B64)
        self.assertEqual(hash(addr), hash((TEST_B32, None)))

    def test_hashWithHostNoPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p')
        self.assertEqual(hash(addr), hash(('spam.i2p', None)))

    def test_hashWithNoHostPort(self):
        addr = I2PAddress(TEST_B64, port=81)
        self.assertEqual(hash(addr), hash((TEST_B32, 81)))

    def test_hashWithHostPort(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', 81)
        self.assertEqual(hash(addr), hash(('spam.i2p', 81)))

    def test_hashWithPortString(self):
        addr = I2PAddress(TEST_B64, 'spam.i2p', '81')
        self.assertEqual(hash(addr), hash(('spam.i2p', 81)))



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.error import ConnectionLost, ConnectionRefusedError
from twisted.python import failure
from twisted.trial import unittest

from txi2p.bob import endpoints
from txi2p.test.util import FakeEndpoint, FakeFactory


connectionLostFailure = failure.Failure(ConnectionLost())
connectionRefusedFailure = failure.Failure(ConnectionRefusedError())



class BOBI2PClientEndpointTestCase(unittest.TestCase):
    """
    Tests for I2P client Endpoint backed by the BOB API.
    """

    def test_bobConnectionFailed(self):
        reactor = object()
        bobEndpoint = FakeEndpoint(failure=connectionRefusedFailure)
        endpoint = endpoints.BOBI2PClientEndpoint(reactor, bobEndpoint, '')
        d = endpoint.connect(None)
        return self.assertFailure(d, ConnectionRefusedError)


    def TODO_test_destination(self):
        reactor = object()
        bobEndpoint = FakeEndpoint()
        endpoint = endpoints.BOBI2PClientEndpoint(reactor, bobEndpoint, 'foo.i2p')
        endpoint.connect(None)
        self.assertEqual(bobEndpoint.transport.value(), 'foo.i2p') # TODO: Fix.


    def TODO_test_clientDataSent(self):
        reactor = object()
        wrappedFac = FakeFactory()
        bobEndpoint = FakeEndpoint()
        endpoint = endpoints.BOBI2PClientEndpoint(reactor, bobEndpoint, '')
        endpoint.connect(wrappedFac)
        bobEndpoint.proto.transport.clear()
        wrappedFac.proto.transport.write('xxxxx')
        self.assertEqual(bobEndpoint.proto.transport.value(), 'xxxxx')



#!/bin/env/python

import os

try:
    from contextlib import suppress
except ImportError:
    from contextlib import contextmanager
    @contextmanager
    def suppress(*exceptions):
        try:
            yield
        except exceptions:
            pass

from distutils.sysconfig import get_python_lib
from subprocess import call


if __name__ == '__main__':
    # chdir to the site-packages directory so the report lists relative paths
    dot_coverage_path = os.path.join(os.getcwd(), '.coverage')
    os.chdir(get_python_lib())
    with suppress(OSError):
        os.remove('.coverage')
    os.symlink(dot_coverage_path, '.coverage')

    # create a report from the coverage data
    if 'TRAVIS' in os.environ:
        rc = call('coveralls')
        # don't fail test if coveralls submission failed
        raise SystemExit(None)
    else:
        rc = call(['coverage', 'report'])
        raise SystemExit(rc)



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from __future__ import print_function
from builtins import range
from builtins import object
import os
from parsley import makeProtocol
from twisted.internet.error import ConnectError, UnknownHostError
from twisted.internet.interfaces import IListeningPort
from twisted.internet.protocol import Protocol
from twisted.python.failure import Failure
from zope.interface import implementer

from txi2p import grammar
from txi2p.address import (
    I2PAddress,
    I2PTunnelTransport,
    I2PServerTunnelProtocol,
)

DEFAULT_INPORT  = 9000
DEFAULT_OUTPORT = 9001


class BOBSender(object):
    def __init__(self, transport):
        self.transport = transport

    def sendClear(self):
        self.transport.write(b'clear\n')

    def sendGetdest(self):
        self.transport.write(b'getdest\n')

    def sendGetkeys(self):
        self.transport.write(b'getkeys\n')

    def sendGetnick(self, tunnelNick):
        msg = 'getnick %s\n' % tunnelNick
        self.transport.write(msg.encode('utf-8'))

    def sendInhost(self, inhost):
        msg = 'inhost %s\n' % inhost
        self.transport.write(msg.encode('utf-8'))

    def sendInport(self, inport):
        msg = 'inport %s\n' % inport
        self.transport.write(msg.encode('utf-8'))

    def sendList(self):
        self.transport.write(b'list\n')

    def sendNewkeys(self):
        self.transport.write(b'newkeys\n')

    def sendOption(self, options={}):
        msg = 'option'
        for key in sorted(options):
            msg += ' %s=%s' % (key, options[key])
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))

    def sendOuthost(self, outhost):
        msg = 'outhost %s\n' % outhost
        self.transport.write(msg.encode('utf-8'))

    def sendOutport(self, outport):
        msg = 'outport %s\n' % outport
        self.transport.write(msg.encode('utf-8'))

    def sendQuiet(self):
        self.transport.write(b'quiet\n')

    def sendQuit(self):
        self.transport.write(b'quit\n')

    def sendSetkeys(self, keys):
        msg = 'setkeys %s\n' % keys
        self.transport.write(msg.encode('utf-8'))

    def sendSetnick(self, tunnelNick):
        msg = 'setnick %s\n' % tunnelNick
        self.transport.write(msg.encode('utf-8'))

    def sendShow(self):
        self.transport.write(b'show\n')

    def sendShowprops(self):
        self.transport.write(b'showprops\n')

    def sendStart(self):
        self.transport.write(b'start\n')

    def sendStatus(self, tunnelNick):
        msg = 'status %s\n' % tunnelNick
        self.transport.write(msg.encode('utf-8'))

    def sendStop(self):
        self.transport.write(b'stop\n')

    def sendVerify(self, key):
        msg = 'verify %s\n' % key
        self.transport.write(msg.encode('utf-8'))

    def sendVisit(self):
        self.transport.write(b'visit\n')


class BOBReceiver(object):
    currentRule = 'State_init'

    def __init__(self, sender):
        self.sender = sender
        self.tunnelExists = False

    def prepareParsing(self, parser):
        # Store the factory for later use
        self.factory = parser.factory

    def finishParsing(self, reason):
        if self.currentRule != 'State_quit':
            self.factory.bobConnectionFailed(reason)

    def initBOB(self, version):
        self.sender.sendList()
        self.currentRule = 'State_list'

    def processTunnelList(self, tunnels):
        if not (hasattr(self.factory, 'tunnelNick') and self.factory.tunnelNick):
            # All tunnels in the same process use the same tunnelNick
            # TODO is using the PID a security risk?
            self.factory.tunnelNick = 'txi2p-%d' % os.getpid()

        used_ports = []
        for i in range(0, len(tunnels)):
            if tunnels[i]['nickname'] == self.factory.tunnelNick:
                # Tunnel already exists, use its settings.
                self.tunnelExists = True
                self.tunnelRunning = tunnels[i]['running']
                self.factory.inhost = tunnels[i]['inhost']
                self.factory.inport = tunnels[i]['inport']
                self.factory.outhost = tunnels[i]['outhost']
                self.factory.outport = tunnels[i]['outport']
                # The tunnel will be removed by the Factory
                # that created it.
                self.factory.removeTunnelWhenFinished = False
                break
            else:
                if tunnels[i]['inport']:
                    used_ports.append(tunnels[i]['inport'])
                if tunnels[i]['outport']:
                    used_ports.append(tunnels[i]['outport'])
        else:
            # This is a new tunnel.
            # Default port offset is at the end of the tunnels list
            offset = 2*(len(tunnels))
            # Find an offset that does not clash
            while (DEFAULT_INPORT + offset) in used_ports or (DEFAULT_OUTPORT + offset) in used_ports:
                offset += 2
            # If the in/outport were not user-configured, set them.
            if not (hasattr(self.factory, 'inport') and self.factory.inport):
                self.factory.inport = DEFAULT_INPORT + offset
            if not (hasattr(self.factory, 'outport') and self.factory.outport):
                self.factory.outport = DEFAULT_OUTPORT + offset

    def getnick(self, success, info):
        if success:
            if self.tunnelRunning:
                self.sender.sendStop()
                self.currentRule = 'State_stop'
            else:
                # Update the local Destination
                self.sender.sendGetdest()
                self.currentRule = 'State_getdest'
        else:
            print('ERROR: %s' % info)

    def stop(self, success, info):
        if success:
            # Update the local Destination
            self.sender.sendGetdest()
            self.currentRule = 'State_getdest'
        else:
            print('stop ERROR: %s' % info)

    def setnick(self, success, info):
        if success:
            # Set the options
            self.sender.sendOption(self.factory.options)
            self.currentRule = 'State_option'
        else:
            print('setnick ERROR: %s' % info)

    def option(self, success, info):
        if success:
            if hasattr(self.factory, 'keypair') and self.factory.keypair: # If a keypair was provided, use it
                self.sender.sendSetkeys(self.factory.keypair)
                self.currentRule = 'State_setkeys'
            else: # Get a new keypair
                self.sender.sendNewkeys()
                self.currentRule = 'State_newkeys'
        else:
            print('option ERROR: %s' % info)

    def setkeys(self, success, info):
        if success:
            # Update the local Destination
            self.sender.sendGetdest()
            self.currentRule = 'State_getdest'
        else:
            print('setkeys ERROR: %s' % info)

    def newkeys(self, success, info):
        if success:
            # Save the new local Destination
            self.factory.localDest = info
            # Get the new keypair
            self.sender.sendGetkeys()
            self.currentRule = 'State_getkeys'
        else:
            print('newkeys ERROR: %s' % info)

    def quit(self, success, info):
        pass


class I2PClientTunnelCreatorBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Set tunnel nickname (and update keypair/localDest state)
                self.sender.sendSetnick(self.factory.tunnelNick)
                self.currentRule = 'State_setnick'
        else:
            print('list ERROR: %s' % info)

    def getdest(self, success, info):
        if success:
            # Save the local Destination
            self.factory.localDest = info
            if self.tunnelExists:
                # Get the keypair
                self.sender.sendGetkeys()
                self.currentRule = 'State_getkeys'
            else:
                self._setInhost()
        else:
            print('getdest ERROR: %s' % info)

    def getkeys(self, success, info):
        if success:
            # Save the keypair
            self.factory.keypair = info
            self._setInhost()
        else:
            print('getkeys ERROR: %s' % info)

    def _setInhost(self):
        self.sender.sendInhost(self.factory.inhost)
        self.currentRule = 'State_inhost'

    def inhost(self, success, info):
        if success:
            self.sender.sendInport(self.factory.inport)
            self.currentRule = 'State_inport'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendInhost(self.factory.inhost)
            else:
                print('inhost ERROR: %s' % info)

    def inport(self, success, info):
        if success:
            self.sender.sendStart()
            self.currentRule = 'State_start'
        else:
            print('inport ERROR: %s' % info)

    def start(self, success, info):
        if success:
            print("Client tunnel started")
            self.factory.i2pTunnelCreated()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            print('start ERROR: %s' % info)


class I2PServerTunnelCreatorBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Set tunnel nickname (and update keypair/localDest state)
                self.sender.sendSetnick(self.factory.tunnelNick)
                self.currentRule = 'State_setnick'
        else:
            print('list ERROR: %s' % info)

    def getdest(self, success, info):
        if success:
            # Save the local Destination
            self.factory.localDest = info
            self._setOuthost()
        else:
            print('getdest ERROR: %s' % info)

    def getkeys(self, success, info):
        if success:
            # Save the keypair
            self.factory.keypair = info
            self._setOuthost()
        else:
            print('getkeys ERROR: %s' % info)

    def _setOuthost(self):
        self.sender.sendOuthost(self.factory.outhost)
        self.currentRule = 'State_outhost'

    def outhost(self, success, info):
        if success:
            self.sender.sendOutport(self.factory.outport)
            self.currentRule = 'State_outport'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendOuthost(self.factory.outhost)
            else:
                print('outhost ERROR: %s' % info)

    def outport(self, success, info):
        if success:
            self.sender.sendStart()
            self.currentRule = 'State_start'
        else:
            print('outport ERROR: %s' % info)

    def start(self, success, info):
        if success:
            print("Server tunnel started")
            self.factory.i2pTunnelCreated()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            print('start ERROR: %s' % info)

class I2PTunnelRemoverBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                # Get tunnel for nickname
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Tunnel already removed
                pass
        else:
            print('list ERROR: %s' % info)

    def getnick(self, success, info):
        if success:
            self.sender.sendStop()
            self.currentRule = 'State_stop'
        else:
            print('getnick ERROR: %s' % info)

    def stop(self, success, info):
        if success:
            self.sender.sendClear()
            self.currentRule = 'State_clear'
        else:
            print('stop ERROR: %s' % info)

    def clear(self, success, info):
        if success:
            print('Tunnel removed')
            self.factory.i2pTunnelRemoved()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendClear()
            else:
                print('clear ERROR: %s ' % info)


# A Protocol for making an I2P client tunnel via BOB
I2PClientTunnelCreatorBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PClientTunnelCreatorBOBReceiver)

# A Protocol for making an I2P server tunnel via BOB
I2PServerTunnelCreatorBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PServerTunnelCreatorBOBReceiver)

# A Protocol for removing a BOB I2P tunnel
I2PTunnelRemoverBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PTunnelRemoverBOBReceiver)


class I2PClientTunnelProtocol(Protocol):
    def __init__(self, wrappedProto, clientAddr, dest):
        self.wrappedProto = wrappedProto
        self._clientAddr = clientAddr
        self.dest = dest
        self._errmsg = None

    def connectionMade(self):
        # Substitute transport for an I2P wrapper
        self.transport = I2PTunnelTransport(self.transport, self._clientAddr,
                                            I2PAddress(self.dest))
        self.isConnected = False
        # First line sent must be the Destination to connect to.
        self.transport.write((self.dest + '\n').encode('utf-8'))
        self.wrappedProto.makeConnection(self.transport)

    def dataReceived(self, data):
        # Check for a successful connection
        if not self.isConnected:
            if data.startswith('ERROR'):
                self._errmsg = data[6:]
                # I2P connection failed
                self.transport.loseConnection()
                return
            else:
                self.isConnected = True

        # Pass all received data to the wrapped Protocol.
        self.wrappedProto.dataReceived(data.encode('utf-8'))

    def connectionLost(self, reason):
        if self._errmsg:
            if self._errmsg.startswith("Can't find destination"):
                reason = Failure(UnknownHostError(string=self._errmsg))
            else:
                reason = Failure(ConnectError(string=self._errmsg))
        self.factory.i2pConnectionLost(self.wrappedProto, reason)


@implementer(IListeningPort)
class I2PListeningPort(object):
    def __init__(self, wrappedPort, factoryWrapper, serverAddr):
        self._wrappedPort = wrappedPort
        self._factoryWrapper = factoryWrapper
        self._serverAddr = serverAddr

    def startListening(self):
        self._wrappedPort.startListening()

    def stopListening(self):
        self._factoryWrapper.stopListening(self._wrappedPort)

    def getHost(self):
        return self._serverAddr



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.error import (
    ConnectBindError,
    ConnectError,
    NoRouteError,
    TCPTimedOutError,
    UnknownHostError,
)


DEFAULT_SIGTYPE = 'EdDSA_SHA512_Ed25519'

RESULT_OK = 'OK'
RESULT_CANT_REACH_PEER = 'CANT_REACH_PEER'
RESULT_DUPLICATED_DEST = 'DUPLICATED_DEST'
RESULT_I2P_ERROR = 'I2P_ERROR'
RESULT_INVALID_KEY = 'INVALID_KEY'
RESULT_KEY_NOT_FOUND = 'KEY_NOT_FOUND'
RESULT_PEER_NOT_FOUND = 'PEER_NOT_FOUND'
RESULT_INVALID_ID = 'INVALID_ID'
RESULT_TIMEOUT = 'TIMEOUT'

samErrorMap = {
    RESULT_CANT_REACH_PEER: NoRouteError,
    RESULT_DUPLICATED_DEST: ConnectBindError,
    RESULT_I2P_ERROR: ConnectError,
    RESULT_INVALID_KEY: ConnectError,
    RESULT_KEY_NOT_FOUND: UnknownHostError,
    RESULT_PEER_NOT_FOUND: NoRouteError,
    RESULT_INVALID_ID: ConnectError,
    RESULT_TIMEOUT: TCPTimedOutError,
}



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
import base64
import hashlib
from twisted.internet.interfaces import IAddress, ITransport
from twisted.internet.protocol import Protocol
from twisted.python.util import FancyEqMixin
from zope.interface import implementer


@implementer(IAddress)
class I2PAddress(FancyEqMixin, object):
    """An :class:`IAddress` that represents the address of an I2P Destination.

    Args:
        destination (str): An I2P Destination string in I2P-style B64 format, or
            an :class:`I2PAddress`. In the latter case, the default host is also
            taken from the provided address.
        host (str): An I2P host string; for example, ``'example.i2p'``.
        port (int): An integer representing the port number.

    Attributes:
        destination (str): An I2P Destination string in I2P-style B64 format.
        host (str): An I2P host string; for example, ``'example.i2p'`` or
            ``'fiftytwocharacters.b32.i2p'``. If looked up, it is guaranteed to
            resolve to ``destination``.
        port (int): An integer representing the port number. Will be ``None`` if
            no port is configured.
    """
    compareAttributes = ('destination', 'port')

    def __init__(self, destination, host=None, port=None):
        if hasattr(destination, 'destination'):
            self.destination = destination.destination
        else:
            self.destination = destination
        self.port = int(port) if port else None

        if host:
            self.host = host
        elif hasattr(destination, 'host'):
            self.host = destination.host
        else:
            raw_key = base64.b64decode(destination.encode('utf-8'), b'-~')
            hash = hashlib.sha256(raw_key)
            base32_hash = base64.b32encode(hash.digest()).decode('utf-8')
            self.host = base32_hash.lower().replace('=', '')+'.b32.i2p'


    def __repr__(self):
        if self.port:
            return '%s(%s, %d)' % (
                self.__class__.__name__, self.host, self.port)
        return '%s(%s)' % (
            self.__class__.__name__, self.host)


    def __hash__(self):
        return hash((self.host, self.port))


@implementer(ITransport)
class I2PTunnelTransport(object):
    def __init__(self, wrappedTransport, localAddr, peerAddr=None, invertTLS=False):
        self.t = wrappedTransport
        self._localAddr = localAddr
        self.peerAddr = peerAddr

        # Workaround for https://tahoe-lafs.org/trac/tahoe-lafs/ticket/2861
        if invertTLS and hasattr(self.t, 'startTLS'):
            import types
            def startTLSWrapper(self, ctx, normal=True):
                self.t.startTLS(ctx, not normal)
            self.startTLS = types.MethodType(startTLSWrapper, self)

    def __getattr__(self, attr):
        return getattr(self.t, attr)

    def getPeer(self):
        return self.peerAddr

    def getHost(self):
        return self._localAddr


class I2PServerTunnelProtocol(Protocol):
    def __init__(self, wrappedProto, serverAddr):
        self.wrappedProto = wrappedProto
        self._serverAddr = serverAddr
        self.peer = None

    def connectionMade(self):
        # Substitute transport for an I2P wrapper
        self.transport = I2PTunnelTransport(self.transport, self._serverAddr)
        self.wrappedProto.makeConnection(self.transport)

    def dataReceived(self, data):
        if self.peer:
            # Pass all other data to the wrapped Protocol.
            self.wrappedProto.dataReceived(data.encode('utf-8'))
        else:
            # First line is the peer's Destination.
            self.setPeer(data)

    def setPeer(self, data):
        self.peer = I2PAddress(data.split('\n')[0])
        self.transport.peerAddr = self.peer

    def connectionLost(self, reason):
        self.wrappedProto.connectionLost(reason)



from __future__ import print_function
from twisted.internet import reactor, defer
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver


class Chat(LineReceiver):

    def __init__(self, users):
        self.users = users
        self.name = None
        self.state = "GETNAME"

    def connectionMade(self):
        self.sendLine("What's your name?")

    def connectionLost(self, reason):
        if self.users.has_key(self.name):
            del self.users[self.name]

    def lineReceived(self, line):
        if self.state == "GETNAME":
            self.handle_GETNAME(line)
        else:
            self.handle_CHAT(line)

    def handle_GETNAME(self, name):
        if self.users.has_key(name):
            self.sendLine("Name taken, please choose another.")
            return
        self.sendLine("Welcome, %s!" % (name,))
        self.name = name
        self.users[name] = self
        self.state = "CHAT"

    def handle_CHAT(self, message):
        message = "<%s> %s" % (self.name, message)
        for name, protocol in self.users.iteritems():
            if protocol != self:
                protocol.sendLine(message)


class ChatFactory(Factory):

    def __init__(self):
        self.users = {} # maps user names to Chat instances

    def buildProtocol(self, addr):
        return Chat(self.users)


def printDest(port):
    # Print out the I2P Destination to copy to the client
    print('This server is listening on:')
    print(port.getHost().destination)
    # Handle Ctl+C
    def shutdown():
        print('Shutting down')
        port.stopListening()
        d = defer.Deferred()
        reactor.callLater(3, d.callback, 1)
        return d
    reactor.addSystemEventTrigger('before', 'shutdown', shutdown)


endpoint = serverFromString(reactor, 'i2p:keypair.chatserver')
d = endpoint.listen(ChatFactory())
d.addCallback(printDest)

reactor.run()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import str
from builtins import object
import functools
from ometa.grammar import OMeta
from ometa.protocol import ParserProtocol
import re
import time
from twisted.internet import reactor
from twisted.internet.interfaces import IProtocolFactory
from twisted.internet.protocol import ClientFactory
from twisted.python.failure import Failure
from zope.interface import implementer

from txi2p import grammar
from txi2p.address import (
    I2PAddress,
    I2PServerTunnelProtocol,
    I2PTunnelTransport,
)
from txi2p.sam import constants as c

KEEPALIVE_TIMEOUT = 2 * 60


def cmpSAM(a, b):
    def normalize(v):
        return [int(x) for x in re.sub(r'(\.0+)*$','', v).split(".")]
    a_n = normalize(a)
    b_n = normalize(b)
    return (a_n > b_n) - (a_n < b_n)

def peerSAM(data):
    peerInfo = data.decode('utf-8').split('\n')[0].split(' ')
    peerOptions = {x: y for x, y in [x.split('=', 1) for x in peerInfo[1:] if x]}
    fromPort = peerOptions['FROM_PORT'] if 'FROM_PORT' in peerOptions else None
    return I2PAddress(peerInfo[0], port=fromPort)


class SAMParserProtocol(ParserProtocol):
    def __init__(self, *args):
        ParserProtocol.__init__(self, *args)

    def dataReceived(self, data):
        """
        Receive and parse some data.

        :param data: A ``bytes`` from Twisted.
        """

        if self._disconnecting:
            return

        if self.receiver.currentRule == 'State_readData':
            # Shortcut for efficiency
            self.receiver.dataReceived(data)
        else:
            # Duplicated from Parsley because it expects a str but Twisted
            # provides a bytes.
            try:
                self._parser.receive(data.decode('utf-8'))
            except Exception:
                self.connectionLost(Failure())
                self.transport.abortConnection()
                return


def makeSAMProtocol(senderFactory, receiverFactory):
    g = OMeta(grammar.samGrammarSource).parseGrammar('Grammar')
    return functools.partial(
        SAMParserProtocol, g, senderFactory, receiverFactory, {})


class SAMSender(object):
    def __init__(self, transport):
        self.transport = transport

    def sendHello(self):
        self.transport.write(b'HELLO VERSION MIN=3.0 MAX=3.2\n')

    def sendNamingLookup(self, name):
        msg = 'NAMING LOOKUP NAME=%s\n' % name
        self.transport.write(msg.encode('utf-8'))

    def sendPing(self, data):
        if data:
            self.transport.write(('PING %s\n' % data).encode('utf-8'))
        else:
            self.transport.write(b'PING\n')

    def sendPong(self, data):
        if data:
            self.transport.write(('PONG %s\n' % data).encode('utf-8'))
        else:
            self.transport.write(b'PONG\n')


class SAMReceiver(object):
    wrappedProto = None
    currentRule = 'State_hello'
    pinger = None
    lastPing = ''
    pingTimeout = None

    def __init__(self, sender):
        self.sender = sender

    def prepareParsing(self, parser):
        # Store the factory for later use
        self.factory = parser.factory
        self.sender.sendHello()

    def wrapProto(self, proto, peerAddress, invertTLS=False):
        self.wrappedProto = proto
        if hasattr(self.factory, 'localPort'):
            localAddress = I2PAddress(self.factory.session.address,
                                      port=self.factory.localPort)
        else:
            localAddress = self.factory.session.address
        self.transportWrapper = I2PTunnelTransport(
            self.sender.transport,
            localAddress, peerAddress,
            invertTLS)
        proto.makeConnection(self.transportWrapper)

    def dataReceived(self, data):
        self.wrappedProto.dataReceived(data)

    def finishParsing(self, reason):
        if self.wrappedProto:
            self.wrappedProto.connectionLost(reason)
        else:
            self.factory.connectionFailed(reason)
        if hasattr(self.factory, 'session'):
            self.factory.session.removeStream(self)

    def hello(self, result, version=None, message=None):
        if result != c.RESULT_OK:
            self.factory.resultNotOK(result, message)
            return
        self.factory.samVersion = version
        self.command()

    def lookupReply(self, result, name, value=None, message=None):
        if result != c.RESULT_OK:
            self.factory.resultNotOK(result, message)
            return
        self.postLookup(value)

    def _sendPing(self):
        self.lastPing = str(time.time())
        self.sender.sendPing(self.lastPing)
        self.pingTimeout = reactor.callLater(KEEPALIVE_TIMEOUT, self.sender.transport.loseConnection)

    def _resetPingTimeout(self):
        if self.pingTimeout:
            self.pingTimeout.cancel()
        self.pinger = reactor.callLater(KEEPALIVE_TIMEOUT, self._sendPing)

    def ping(self, data):
        self.sender.sendPong(data)
        self._resetPingTimeout()

    def pong(self, data):
        if (data == str(self.lastPing)):
            self._resetPingTimeout()

    def startPinging(self):
        self.pinger = reactor.callLater(KEEPALIVE_TIMEOUT, self._sendPing)
        self.currentRule = 'State_keepalive'

    def stopPinging(self):
        if self.pinger and self.pinger.active():
            self.pinger.cancel()
        if self.pingTimeout and self.pingTimeout.active():
            self.pingTimeout.cancel()

class SAMFactory(ClientFactory):
    currentCandidate = None
    canceled = False

    def _cancel(self, d):
        self.currentCandidate.sender.transport.abortConnection()
        self.canceled = True

    def buildProtocol(self, addr):
        proto = self.protocol()
        proto.factory = self
        self.currentCandidate = proto
        return proto

    def connectionFailed(self, reason):
        if not self.canceled and not self.deferred.called:
            self.deferred.errback(reason)

    # This method is not called if an endpoint deferred errbacks
    def clientConnectionFailed(self, connector, reason):
        self.connectionFailed(reason)

    def resultNotOK(self, result, message):
        raise c.samErrorMap.get(result)(string=(message if message else result))


class SAMI2PServerTunnelProtocol(I2PServerTunnelProtocol):
    def setPeer(self, data):
        self.peer = peerSAM(data)
        self.transport.peerAddr = self.peer


@implementer(IProtocolFactory)
class I2PFactoryWrapper(object):
    protocol = SAMI2PServerTunnelProtocol

    def __init__(self, wrappedFactory, serverAddr):
        self.w = wrappedFactory
        self.serverAddr = serverAddr

    def buildProtocol(self, addr):
        wrappedProto = self.w.buildProtocol(addr)
        proto = self.protocol(wrappedProto, self.serverAddr)
        proto.factory = self
        return proto

    def __getattr__(self, attr):
        return getattr(self.w, attr)



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

import unittest

from parsley import makeGrammar, ParseError

from txi2p.grammar import bobGrammarSource, samGrammarSource


bobGrammar = makeGrammar(bobGrammarSource, {})
samGrammar = makeGrammar(samGrammarSource, {})

def stringParserFromRule(grammar, rule):
    def parseString(s):
        return getattr(grammar(s), rule)()
    return parseString


class TestBOBGrammar(unittest.TestCase):
    def _test(self, rule, data, expected):
        parse = stringParserFromRule(bobGrammar, rule)
        result = parse(data)
        self.assertEqual(result, expected)

    def test_BOB_clear(self):
        self._test('BOB_clear', 'OK cleared\n', (True, 'cleared'))
        self._test('BOB_clear', 'ERROR tunnel is active\n', (False, 'tunnel is active'))

    def test_BOB_getdest(self):
        self._test('BOB_getdest', 'OK spam\n', (True, 'spam'))

    def test_BOB_getkeys(self):
        self._test('BOB_getkeys', 'OK spameggs\n', (True, 'spameggs'))

    def test_BOB_list(self):
        spam = {
            'nickname': 'spam',
            'starting': False,
            'running': True,
            'stopping': False,
            'keys': True,
            'quiet': False,
            'inport': 12345,
            'inhost': 'localhost',
            'outport': 23456,
            'outhost': 'localhost'
        }
        eggs = {
            'nickname': 'eggs',
            'starting': False,
            'running': False,
            'stopping': False,
            'keys': True,
            'quiet': False,
            'inport': None,
            'inhost': 'localhost',
            'outport': None,
            'outhost': 'localhost'
        }

        self._test('BOB_list', 'OK Listing done\n', (True, 'Listing done', []))
        self._test('BOB_list', 'DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nDATA NICKNAME: eggs STARTING: false RUNNING: false STOPPING: false KEYS: true QUIET: false INPORT: not_set INHOST: localhost OUTPORT: not_set OUTHOST: localhost\nOK Listing done\n', (True, 'Listing done', [spam, eggs]))
        self._test('BOB_list', 'ERROR ni!\n', (False, 'ni!', []))


class TestSAMGrammar(unittest.TestCase):
    def _test(self, rule, data, expected):
        parse = stringParserFromRule(samGrammar, rule)
        result = parse(data)
        self.assertEqual(result, expected)

    def test_SAM_hello(self):
        self._test('SAM_hello', 'HELLO REPLY RESULT=OK VERSION=3.1\n',
                   {'result': 'OK', 'version': '3.1'})
        self._test('SAM_hello', 'HELLO REPLY RESULT=NOVERSION\n', {'result': 'NOVERSION'})
        self._test('SAM_hello',
                   'HELLO REPLY RESULT=I2P_ERROR MESSAGE="Something failed"\n',
                   {'result': 'I2P_ERROR', 'message': 'Something failed'})

    def test_SAM_session_status(self):
        self._test('SAM_session_status',
                   'SESSION STATUS RESULT=OK DESTINATION=privkey\n',
                   {'result': 'OK', 'destination': 'privkey'})
        self._test('SAM_session_status',
                   'SESSION STATUS RESULT=DUPLICATED_ID\n',
                   {'result': 'DUPLICATED_ID'})

    def test_SAM_stream_status(self):
        self._test('SAM_stream_status', 'STREAM STATUS RESULT=OK\n', {'result': 'OK'})
        self._test('SAM_stream_status',
                   'STREAM STATUS RESULT=CANT_REACH_PEER MESSAGE="Can\'t reach peer"\n',
                   {'result': 'CANT_REACH_PEER', 'message': 'Can\'t reach peer'})

    def test_SAM_naming_reply(self):
        self._test('SAM_naming_reply',
                   'NAMING REPLY RESULT=OK NAME=name VALUE=dest\n',
                   {'result': 'OK', 'name': 'name', 'value': 'dest'})
        self._test('SAM_naming_reply',
                   'NAMING REPLY RESULT=KEY_NOT_FOUND\n',
                   {'result': 'KEY_NOT_FOUND'})

    def test_SAM_dest_reply(self):
        self._test('SAM_dest_reply',
                   'DEST REPLY PUB=foo PRIV=foobar\n',
                   {'pub': 'foo', 'priv': 'foobar'})

    def test_SAM_ping(self):
        self._test('SAM_ping', 'PING\n', None)
        self._test('SAM_ping', 'PING 1234567890\n', '1234567890')

    def test_SAM_pong(self):
        self._test('SAM_pong', 'PONG\n', None)
        self._test('SAM_pong', 'PONG 1234567890\n', '1234567890')



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import range
from builtins import object
from twisted.internet.defer import Deferred
from twisted.internet.error import ConnectError, UnknownHostError
from twisted.internet.interfaces import IListeningPort
from zope.interface import implementer

from txi2p.address import I2PAddress
from txi2p.sam import constants as c
from txi2p.sam.base import (
    cmpSAM,
    peerSAM,
    makeSAMProtocol,
    SAMSender,
    SAMReceiver,
    SAMFactory,
)


class StreamConnectSender(SAMSender):
    def sendStreamConnect(self, id, destination, port=None, localPort=None):
        msg = 'STREAM CONNECT'
        msg += ' ID=%s' % id
        msg += ' DESTINATION=%s' % destination
        msg += ' SILENT=false'
        if port:
            msg += ' TO_PORT=%d' % port
        if localPort:
            msg += ' FROM_PORT=%d' % localPort
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))


class StreamConnectReceiver(SAMReceiver):
    def command(self):
        if self.factory.dest:
            self.doConnect()
        else:
            self.sender.sendNamingLookup(self.factory.host)
            self.currentRule = 'State_naming'

    def postLookup(self, dest):
        self.factory.dest = dest
        self.doConnect()

    def doConnect(self):
        self.sender.sendStreamConnect(
            self.factory.session.id, self.factory.dest,
            self.factory.port, self.factory.localPort)
        self.currentRule = 'State_connect'

    def connect(self, result, message=None):
        if result != c.RESULT_OK:
            self.factory.resultNotOK(result, message)
            return

        self.factory.streamConnectionEstablished(self)
        self.currentRule = 'State_readData'


StreamConnectProtocol = makeSAMProtocol(
    StreamConnectSender,
    StreamConnectReceiver)


class StreamConnectFactory(SAMFactory):
    protocol = StreamConnectProtocol

    def __init__(self, clientFactory, session, host, dest, port=None, localPort=None):
        self._clientFactory = clientFactory
        self.session = session
        self.host = host
        self.dest = dest
        self.port = port
        self.localPort = localPort
        self.deferred = Deferred(self._cancel);

    def streamConnectionEstablished(self, streamProto):
        self.session.addStream(streamProto)
        peerAddress = I2PAddress(self.dest, self.host, self.port)
        proto = self._clientFactory.buildProtocol(peerAddress)
        if proto is None:
            self.deferred.cancel()
            return
        streamProto.wrapProto(proto, peerAddress)
        self.deferred.callback(proto)


class StreamAcceptSender(SAMSender):
    def sendStreamAccept(self, id):
        msg = 'STREAM ACCEPT'
        msg += ' ID=%s' % id
        msg += ' SILENT=false'
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))


class StreamAcceptReceiver(SAMReceiver):
    peer = None
    initialData = b''

    def command(self):
        self.sender.sendStreamAccept(
            self.factory.session.id)
        self.currentRule = 'State_accept'

    def accept(self, result, message=None):
        if result != c.RESULT_OK:
            self.factory.resultNotOK(result, message)
            return
        self.factory.streamAcceptEstablished(self)
        self.currentRule = 'State_readData'

    def dataReceived(self, data):
        if self.peer:
            # Pass all other data to the wrapped Protocol.
            self.wrappedProto.dataReceived(data)
        else:
            self.initialData += data
            if b'\n' in self.initialData:
                # First line is the peer's Destination.
                data, self.initialData = self.initialData.split(b'\n', 1)
                self.peer = peerSAM(data)
                # Create the wrapped Protocol...
                self.factory.streamAcceptIncoming(self)
                # ... and pass through any initial data.
                if self.initialData:
                    data, self.initialData = self.initialData, None
                    self.wrappedProto.dataReceived(data)


StreamAcceptProtocol = makeSAMProtocol(
    StreamAcceptSender,
    StreamAcceptReceiver)


class StreamAcceptFactory(SAMFactory):
    protocol = StreamAcceptProtocol

    def __init__(self, clientFactory, session, listeningPort):
        self._clientFactory = clientFactory
        self.session = session
        self.listeningPort = listeningPort
        self.deferred = Deferred(self._cancel);

    def streamAcceptEstablished(self, streamProto):
        self.session.addStream(streamProto)
        self.listeningPort.addAccept(streamProto)

    def streamAcceptIncoming(self, streamProto):
        self.listeningPort.removeAccept(streamProto)
        proto = self._clientFactory.buildProtocol(streamProto.peer)
        if proto is None:
            self.deferred.cancel()
            return
        streamProto.wrapProto(proto, streamProto.peer, True)


@implementer(IListeningPort)
class StreamAcceptPort(object):
    def __init__(self, session, factory):
        self.session = session
        self.factory = StreamAcceptFactory(factory, session, self)
        self.accepts = []

    def startListening(self):
        if cmpSAM(self.session.samVersion, '3.2') >= 0:
            active = 8
        else:
            active = 1
        for i in range(0, active):
            self.openAccept()

    def stopListening(self):
        for pending in self.accepts:
            pending.sender.transport.loseConnection()
        self.accepts = []

    def openAccept(self):
        self.session.samEndpoint.connect(self.factory)

    def addAccept(self, proto):
        self.accepts.append(proto)

    def removeAccept(self, proto):
        self.accepts.remove(proto)
        self.openAccept()

    def getHost(self):
        return self.session.address


class StreamForwardSender(SAMSender):
    def sendStreamForward(self, id, port, host=None):
        msg = 'STREAM FORWARD'
        msg += ' ID=%s' % id
        msg += ' PORT=%s' % port
        if host:
            msg += ' HOST=%s' % host
        msg += ' SILENT=false'
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))


class StreamForwardReceiver(SAMReceiver):
    def command(self):
        self.sender.sendStreamForward(
            self.factory.session.id,
            self.factory.forwardPort)
        self.currentRule = 'State_forward'

    def forward(self, result, message=None):
        if result != c.RESULT_OK:
            self.factory.resultNotOK(result, message)
            return
        self.factory.streamForwardEstablished(self)


StreamForwardProtocol = makeSAMProtocol(
    StreamForwardSender,
    StreamForwardReceiver)


class StreamForwardFactory(SAMFactory):
    protocol = StreamForwardProtocol

    def __init__(self, session, forwardPort):
        self.session = session
        self.forwardPort = forwardPort
        self.deferred = Deferred(self._cancel);

    def streamForwardEstablished(self, forwardingProto):
        self.session.addStream(forwardingProto)
        self.deferred.callback(forwardingProto)


@implementer(IListeningPort)
class StreamForwardPort(object):
    def __init__(self, listeningPort, forwardingProto, serverAddr):
        self._listeningPort = listeningPort
        self._forwardingProto = forwardingProto
        self._serverAddr = serverAddr

    def startListening(self):
        self._listeningPort.startListening()

    def stopListening(self):
        self._listeningPort.stopListening()
        self._forwardingProto.sender.transport.loseConnection()

    def getHost(self):
        return self._serverAddr



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from txi2p.plugins import I2PClientParser, I2PServerParser


i2pClientEndpointParser = I2PClientParser()
i2pServerEndpointParser = I2PServerParser()



from __future__ import print_function
from sys import stdout
from twisted.internet import reactor
from twisted.internet.endpoints import clientFromString
from twisted.internet.protocol import ClientFactory, Protocol


class Eepsite(Protocol):
    def connectionMade(self):
        print('Connection made, sending eepsite request')
        self.transport.write(b'GET / HTTP/1.1\r\n\r\n')

    def dataReceived(self, data):
        stdout.write(data)

    def connectionLost(self, reason):
        print('Lost connection. Reason:', reason)
        reactor.stop()


class EepsiteFactory(ClientFactory):
    protocol = Eepsite


endpoint = clientFromString(reactor, 'i2p:stats.i2p:81')
d = endpoint.connect(EepsiteFactory())
def printProto(proto):
    print('My address:')
    print(proto.transport.getHost())
    print('Eepsite address:')
    print(proto.transport.getPeer())
def printErr(err):
    print(err)
    reactor.stop()
d.addCallbacks(printProto, printErr)

reactor.run()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from setuptools import setup
from os import path
import sys


here = path.abspath(path.dirname(__file__))

def readme():
    with open(path.join(here, 'README.rst')) as f:
        return f.read()

install_requires = [
    'Parsley>=1.2',
]

# future is only a requirement for Py2
# This will not work on Py3 if any of the 14 standard library modules listed
# here get used later on:
# http://python-future.org/standard_library_imports.html#list-standard-library-refactored
if sys.version_info[0] < 3:
    install_requires.append('future>=0.14.0')
    install_requires.append('Twisted>=10.1')
else:
    install_requires.append('Twisted>=15.4')

setup(
    name='txi2p-tahoe',
    description='I2P bindings for Twisted',
    long_description=readme(),
    author='str4d',
    author_email='str4d@i2pmail.org',
    url='https://github.com/str4d/txi2p',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet',
    ],
    license='ISC',
    install_requires=install_requires,
    packages=[
        'txi2p',
        'txi2p.bob',
        'txi2p.sam',
        'txi2p.test',
        'txi2p.bob.test',
        'txi2p.sam.test',
        'twisted.plugins',
    ],
    extras_require={
        "test": [
            "mock; python_version < '3.0'",
        ],
    },
)

# Make Twisted regenerate the dropin.cache, if possible.  This is necessary
# because in a site-wide install, dropin.cache cannot be rewritten by
# normal users.
try:
    from twisted.plugin import IPlugin, getPlugins
except ImportError:
    pass
else:
    list(getPlugins(IPlugin))



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
import os
from twisted.internet.error import UnknownHostError
from twisted.internet.protocol import ClientFactory
from twisted.python.failure import Failure
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.bob.protocol import (I2PClientTunnelCreatorBOBClient,
                                I2PServerTunnelCreatorBOBClient,
                                I2PTunnelRemoverBOBClient,
                                I2PClientTunnelProtocol,
                                I2PServerTunnelProtocol,
                                DEFAULT_INPORT, DEFAULT_OUTPORT)

TEST_B64 = "2wDRF5nDfeTNgM4X-TI5xEk3R-WiaTABvkMQ2eYpvEzayUZQJgr9E2T6Y2m9HHn3xHYGEOg-RLisjW9AubTaUTx-v66AsEEtv745qPcuWuV1SP~w1bdzYEn8MSoK7Zh4mwHBg1uHq8z17TUNvWz19q76vHNth-2PDuBToD7ySBn3cGBFDUU83wJJXPD6OueLY8yosWWtksk7WZk60~6z~nVePPSEY8JDry3myLDe11szAVER4A8eX1sFpw247cXGGJK9wQhV-TXFj~m76GPVcFKh7u79zwTwZnZ1GXXKqqyRoj1c4-U69CvvJsQRLmdLFwFEpRkxwV8z6LIFclYJk443YpTnPXC7vNdFOzqqS4FLR1ra~DNfN5foMtR2~2VxuR5m2dYiOS6GzHDxA4acJJSGqnasJjcEIFNVSQKxMnFu9PvGLNJHZ83EraHCErENcOGkPlnVgcJCtPGNGiirwCbBz38jE0lfjkrNrWabc6uWeU559CobG8F8KUDx1irpAAAA"


class BOBProtoTestMixin(object):
    def makeProto(self, *a, **kw):
        protoClass = kw.pop('_protoClass', self.protocol)
        fac = ClientFactory(*a, **kw)
        fac.protocol = protoClass
        fac.options = {}
        def raise_(reason):
            raise reason.value
        fac.bobConnectionFailed = lambda reason: raise_(reason)
        proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return fac, proto

    def test_initBOBListsTunnels(self):
        fac, proto = self.makeProto()
        proto.dataReceived('BOB 00.00.10\nOK\n')
        self.assertEqual(proto.transport.value(), b'list\n')

    def TODO_test_quitDoesNotErrback(self):
        fac, proto = self.makeProto()
        # Shortcut to end of BOB protocol
        proto.receiver.currentRule = 'State_quit'
        proto._parser._setupInterp()
        proto.dataReceived('OK Bye!\n')
        self.assertTrue(False, 'TODO: Test something') # TODO: Test something

class BOBTunnelCreationMixin(BOBProtoTestMixin):
    def test_defaultInportSelected(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertTrue(hasattr(fac, 'inport'))
        self.assertEqual(fac.inport, DEFAULT_INPORT)

    def test_higherInportSelectedWhenDefaultBusy(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: test STARTING: false RUNNING: false STOPPING: false KEYS: false QUIET: false INPORT: 9000 INHOST: localhost OUTPORT: not_set OUTHOST: localhost\nOK Listing done\n')
        self.assertTrue(hasattr(fac, 'inport'))
        self.assertEqual(fac.inport, DEFAULT_INPORT + 2)

    def test_existingInhostAndInportSelectedForExistingTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        fac.inport = 1234
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: false STOPPING: false KEYS: false QUIET: false INPORT: 2345 INHOST: localhost OUTPORT: not_set OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(fac.inhost, 'localhost')
        self.assertEqual(fac.inport, 2345)

    def test_defaultOutportSelected(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertTrue(hasattr(fac, 'outport'))
        self.assertEqual(fac.outport, DEFAULT_OUTPORT)

    def test_higherOutportSelectedWhenDefaultBusy(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: test STARTING: false RUNNING: false STOPPING: false KEYS: false QUIET: false INPORT: not_set INHOST: localhost OUTPORT: 9001 OUTHOST: localhost\nOK Listing done\n')
        self.assertTrue(hasattr(fac, 'outport'))
        self.assertEqual(fac.outport, DEFAULT_OUTPORT + 2)

    def test_existingOuthostAndOutportSelectedForExistingTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        fac.outport = 1234
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: false STOPPING: false KEYS: false QUIET: false INPORT: not_set INHOST: localhost OUTPORT: 2345 OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(fac.outhost, 'localhost')
        self.assertEqual(fac.outport, 2345)

    def test_defaultNickSetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = None
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value().decode('utf-8'), 'setnick txi2p-%d\n' % os.getpid())

    def test_newNickSetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value(), b'setnick spam\n')

    def test_nickSetSendsOptions(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.options = {'foo': 'bar', 'spam': 'eggs'}
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'option foo=bar spam=eggs\n')

    def test_nickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        self.assertEqual(proto.transport.value(), b'setkeys eggs\n')

    def test_destFetchedAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'getdest\n')

    def test_nickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        self.assertEqual(proto.transport.value(), b'newkeys\n')

    def test_keypairFetchedAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        self.assertEqual(proto.transport.value(), b'getkeys\n')

    def test_existingNickGetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(proto.transport.value(), b'getnick spam\n')

    def test_stopRequestedForRunningTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'stop\n')

    def test_stopNotRequestedForStoppedTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: false STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertNotEqual(proto.transport.value(), b'stop\n') # TODO: Refactor to test against what is actually expected

    def test_quitRequestedAfterStart(self):
        fac, proto = self.makeProto()
        called = []
        fac.i2pTunnelCreated = lambda: called.append(True)
        # Shortcut
        proto.receiver.currentRule = 'State_start'
        proto._parser._setupInterp()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual((called, proto.transport.value()), ([True], b'quit\n'))

class TestI2PClientTunnelCreatorBOBClient(BOBTunnelCreationMixin, unittest.TestCase):
    protocol = I2PClientTunnelCreatorBOBClient

    def test_inhostRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_inhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), b'inhost camelot\n')

    def test_inhostRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_inhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), b'inhost camelot\n')

    def test_inhostSetAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The Destination
        self.assertEqual(proto.transport.value(), b'inhost camelot\n')

    def test_inhostSetAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        self.assertEqual(proto.transport.value(), b'inhost camelot\n')

    def test_defaultInportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value().decode('utf-8'), 'inport %d\n' % DEFAULT_INPORT)

    def test_inportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        fac.inport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'inport 1234\n')

    def test_startRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        fac.inport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'start\n')


class TestI2PServerTunnelCreatorBOBClient(BOBTunnelCreationMixin, unittest.TestCase):
    protocol = I2PServerTunnelCreatorBOBClient

    def test_outhostRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_outhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), b'outhost camelot\n')

    def test_outhostRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_outhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), b'outhost camelot\n')

    def test_outhostSetAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The Destination
        self.assertEqual(proto.transport.value(), b'outhost camelot\n')

    def test_outhostSetAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        self.assertEqual(proto.transport.value(), b'outhost camelot\n')

    def test_defaultOutportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value().decode('utf-8'), 'outport %d\n' % DEFAULT_OUTPORT)

    def test_outportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        fac.outport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'outport 1234\n')

    def test_startRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        fac.outport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418 options set\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'start\n')


class TestI2PTunnelRemoverBOBClient(BOBProtoTestMixin, unittest.TestCase):
    protocol = I2PTunnelRemoverBOBClient

    def test_noTunnelWithNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value(), b'')

    def test_tunnelExistsGetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(proto.transport.value(), b'getnick spam\n')

    def test_stopRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'stop\n')

    def test_clearRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'clear\n')

    def test_clearRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        # Shortcut
        proto.receiver.currentRule = 'State_clear'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), b'clear\n')

    def test_clearRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        # Shortcut
        proto.receiver.currentRule = 'State_clear'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), b'clear\n')

    def test_quitRequestedAfterClearSuccess(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.i2pTunnelRemoved = lambda: None
        # Shortcut
        proto.receiver.currentRule = 'State_clear'
        proto._parser._setupInterp()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), b'quit\n')


class FakeDisconnectingFactory(object):
    def i2pConnectionLost(self, wrappedProto, reason):
        wrappedProto.connectionLost(reason)

class TestI2PClientTunnelProtocol(unittest.TestCase):
    def makeProto(self):
        wrappedProto = proto_helpers.AccumulatingProtocol()
        proto = I2PClientTunnelProtocol(wrappedProto, None, TEST_B64)
        proto.factory = FakeDisconnectingFactory()
        transport = proto_helpers.StringTransportWithDisconnection()
        transport.abortConnection = lambda: None
        transport.protocol = proto
        proto.makeConnection(transport)
        return proto

    def test_destRequested(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport.value().decode('utf-8'), '%s\n' % TEST_B64)

    def test_wrappedProtoConnectionMade(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport, proto.wrappedProto.transport)

    def test_connectionFailed(self):
        proto = self.makeProto()
        proto.dataReceived("ERROR Can't find destination: spam.i2p")
        self.assertEqual(proto.wrappedProto.closed, 1)

        expected = UnknownHostError(string="Can't find destination: spam.i2p")
        got = proto.wrappedProto.closedReason.value
        self.assertEqual(type(expected), type(got))
        self.assertEqual(expected.args, got.args)

    def test_dataPassed(self):
        proto = self.makeProto()
        proto.dataReceived('shrubbery')
        self.assertEqual(proto.wrappedProto.data, b'shrubbery')


class TestI2PServerTunnelProtocol(unittest.TestCase):
    def makeProto(self):
        wrappedProto = proto_helpers.AccumulatingProtocol()
        proto = I2PServerTunnelProtocol(wrappedProto, None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return proto

    def test_wrappedProtoConnectionMade(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport, proto.wrappedProto.transport)

    def test_peerDestStored(self):
        proto = self.makeProto()
        proto.dataReceived('%s\n' % TEST_B64)
        self.assertEqual(proto.peer.destination, TEST_B64)

    def test_dataAfterPeerDestPassed(self):
        proto = self.makeProto()
        proto.dataReceived('%s\n' % TEST_B64)
        proto.dataReceived('shrubbery')
        self.assertEqual(proto.wrappedProto.data, b'shrubbery')



from twisted.internet.endpoints import clientFromString

from txi2p import sam


DEFAULT_ENDPOINT = {
    'BOB': 'tcp:127.0.0.1:2827',
    'SAM': 'tcp:127.0.0.1:7656',
    }

DEFAULT_API = 'SAM'

def getApi(api, apiEndpoint, apiDict):
    if not api:
        if apiEndpoint:
            raise ValueError('api must be specified if apiEndpoint is given')
        else:
            api = DEFAULT_API

    if api not in apiDict:
        raise ValueError('Specified I2P API is invalid or unsupported')

    if not apiEndpoint:
        apiEndpoint = DEFAULT_ENDPOINT[api]

    return (api, apiEndpoint)


_apiTesters = {
    'SAM': sam.testAPI,
}

def testAPI(reactor, api=None, apiEndpoint=None):
    """Test whether an API is reachable.

    The function returns a :class:`twisted.internet.defer.Deferred`; register
    callbacks to receive the return value or errors.

    Args:
        reactor: The API endpoint will be constructed with this reactor.
        api (str): The API to test.
        apiEndpoint (str): An endpoint string that may connect to the API.
            Alternatively, the caller can directly provide an
            :class:`twisted.internet.interfaces.IStreamClientEndpoint`, and the
            ``reactor`` will be ignored.

    Returns:
        True if the API is reachable.

    Raises:
        ValueError: if the API doesn't support this method.
    """
    api, apiEndpoint = getApi(api, apiEndpoint, _apiTesters)
    if isinstance(apiEndpoint, str):
        apiEndpoint = clientFromString(reactor, apiEndpoint)
    return _apiTesters[api](apiEndpoint)


_apiGenerators = {
    'SAM': sam.generateDestination,
}

def generateDestination(reactor, keyfile, api=None, apiEndpoint=None):
    """Generate a new I2P Destination.

    The function returns a :class:`twisted.internet.defer.Deferred`; register
    callbacks to receive the return value or errors.

    Args:
        reactor: The API endpoint will be constructed with this reactor.
        keyfile (str): Path to a local file where the keypair for the new
            Destination should be stored.
        api (str): The API to use.
        apiEndpoint (str): An endpoint string that will connect to the API.
            Alternatively, the caller can directly provide an
            :class:`twisted.internet.interfaces.IStreamClientEndpoint`, and the
            ``reactor`` will be ignored.

    Returns:
        txi2p.I2PAddress: The new Destination. Once this is received via the
        Deferred callback, the ``keyfile`` will have been written.

    Raises:
        ValueError: if the API doesn't support this method.
        ValueError: if the ``keyfile`` already exists.
        IOError: if the ``keyfile`` write fails.
    """
    api, apiEndpoint = getApi(api, apiEndpoint, _apiGenerators)
    if isinstance(apiEndpoint, str):
        apiEndpoint = clientFromString(reactor, apiEndpoint)
    return _apiGenerators[api](keyfile, apiEndpoint)
